# molmass.py

# Copyright (c) 1990-2024, Christoph Gohlke
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Molecular mass calculations.

Molmass is a Python library, console script, and web application to calculate
the molecular mass (average, nominal, and isotopic pure), the elemental
composition, and the mass distribution spectrum of a molecule given by its
chemical formula, relative element weights, or sequence.

Calculations are based on the isotopic composition of the elements. Mass
deficiency due to chemical bonding is not taken into account.

The library includes a database of physicochemical and descriptive properties
of the chemical elements.

:Author: `Christoph Gohlke <https://www.cgohlke.com>`_
:License: BSD 3-Clause
:Version: 2024.5.10
:DOI: `10.5281/zenodo.7135495 <https://doi.org/10.5281/zenodo.7135495>`_

Quickstart
----------

Install the molmass package and all dependencies from the
`Python Package Index <https://pypi.org/project/molmass/>`_::

    python -m pip install -U molmass[all]

Print the console script usage::

    python -m molmass --help

Run the web application::

    python -m molmass --web

The molmass library is documented via docstrings.

See `Examples`_ for using the programming interface.

Source code and support are available on
`GitHub <https://github.com/cgohlke/molmass>`_.

Requirements
------------

This revision was tested with the following requirements and dependencies
(other versions may work):

- `CPython <https://www.python.org>`_ 3.9.13, 3.10.11, 3.11.9, 3.12.3
- `Flask <https://pypi.org/project/Flask/>`_ 3.0.3 (optional)
- `Pandas <https://pypi.org/project/pandas/>`_ 2.2.2 (optional)
- `wxPython <https://pypi.org/project/wxPython/>`_ 4.2.1 (optional)

Revisions
---------

2024.5.10

- Add options to disable parsing groups, oligos, fractions, arithmetic (#14).
- Add Formula.expanded property.

2023.8.30

- Fix linting issues.
- Add py.typed marker.
- Remove support for Python 3.8.

2023.4.10

- Support rdkit-style ionic charges (#11, #12).
- Enable multiplication without addition in from_string.

2022.12.9

- Fix split_charge formula with trailing ]] (#11).

2022.10.18

- Several breaking changes.
- Add experimental support for ion charges (#5).
- Change Element, Isotope, and Particle to dataclass (breaking).
- Change types of Spectrum and Composition (breaking).
- Add functions to export Spectrum and Composition as Pandas DataFrames.
- Replace lazyattr with functools.cached_property.
- Rename molmass_web to web (breaking).
- Change output of web application (breaking).
- Run web application using Flask if installed.
- Add options to specify URL of web application and not opening web browser.
- Convert to Google style docstrings.
- Add type hints.
- Remove support for Python 3.7.

2021.6.18

- Add Particle types to elements (#5).
- Fix molmass_web failure on WSL2 (#9).
- Fix elements_gui layout issue.
- Remove support for Python 3.6.

2020.6.10

- Fix elements_gui symbol size on WSL2.
- Support wxPython 4.1.

2020.1.1

- Update elements atomic weights and isotopic compositions from NIST.
- Move element descriptions into separate module.
- Remove support for Python 2.7 and 3.5.

2018.8.15

- Move modules into molmass package.

2018.5.29

- Add option to start web interface from console.
- Separate styles from content and use CSS flex layout in molmass_web.

2018.5.25

- Style and docstring fixes.
- Make from_fractions output deterministic.
- Accept Flask request.args in molmass_web.
- Style and template changes in molmass_web.

2016.2.25

- Fix some elements ionization energies.

2005.x.x

- Initial release.

Examples
--------

Calculate the molecular mass, elemental composition, and mass distribution of
a molecule from its chemical formula:

>>> from molmass import Formula
>>> f = Formula('C8H10N4O2')  # Caffeine
>>> f
Formula('C8H10N4O2')
>>> f.formula  # hill notation
'C8H10N4O2'
>>> f.empirical
'C4H5N2O'
>>> f.mass  # average mass
194.1909...
>>> f.nominal_mass  # == f.isotope.massnumber
194
>>> f.monoisotopic_mass  # == f.isotope.mass
194.0803...
>>> f.atoms
24
>>> f.charge
0
>>> f.composition().dataframe()
         Count  Relative mass  Fraction
Element...
C            8      96.085920  0.494801
H           10      10.079410  0.051905
N            4      56.026812  0.288514
O            2      31.998810  0.164780
>>> f.spectrum(min_intensity=0.01).dataframe()
             Relative mass  Fraction  Intensity %         m/z
Mass number...
194             194.080376  0.898828   100.000000  194.080376
195             195.082873  0.092625    10.305100  195.082873
196             196.084968  0.008022     0.892492  196.084968
197             197.087214  0.000500     0.055681  197.087214

Access physicochemical and descriptive properties of the chemical elements:

>>> from molmass import ELEMENTS, Element
>>> e = ELEMENTS['C']
>>> e
Element(
    6, 'C', 'Carbon',
    group=14, period=2, block='p', series=1,
    mass=12.01074, eleneg=2.55, eleaffin=1.262118,
    covrad=0.77, atmrad=0.91, vdwrad=1.7,
    tboil=5100.0, tmelt=3825.0, density=3.51,
    eleconfig='[He] 2s2 2p2',
    oxistates='4*, 2, -4*',
    ionenergy=(
        11.2603, 24.383, 47.877, 64.492, 392.077,
        489.981,
    ),
    isotopes={
        12: Isotope(12.0, 0.9893, 12),
        13: Isotope(13.00335483507, 0.0107, 13),
    },
)
>>> e.number
6
>>> e.symbol
'C'
>>> e.name
'Carbon'
>>> e.description
'Carbon is a member of group 14 of the periodic table...'
>>> e.eleconfig
'[He] 2s2 2p2'
>>> e.eleconfig_dict
{(1, 's'): 2, (2, 's'): 2, (2, 'p'): 2}
>>> str(ELEMENTS[6])
'Carbon'
>>> len(ELEMENTS)
109
>>> sum(e.mass for e in ELEMENTS)
14693.181589001...
>>> for e in ELEMENTS:
...     e.validate()
...     e = eval(repr(e))
...

"""

from __future__ import annotations

__version__ = '2024.5.10'

__all__ = [
    'Composition',
    'CompositionItem',
    'Formula',
    'FormulaError',
    'Spectrum',
    'SpectrumEntry',
    'analyze',
    'format_charge',
    'from_elements',
    'from_fractions',
    'from_oligo',
    'from_peptide',
    'from_sequence',
    'from_string',
    'hill_sorted',
    'join_charge',
    'main',
    'split_charge',
    'test',
    'AMINOACIDS',
    'DEOXYNUCLEOTIDES',
    'GROUPS',
    'NUCLEOTIDES',
    'PREPROCESSORS',
]

import copy
import math
import re
import sys
from dataclasses import dataclass
from functools import cached_property, reduce
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Iterator, Sequence

    import pandas

try:
    from .elements import ELECTRON, ELEMENTS, Isotope
except ImportError:
    from elements import ELECTRON, ELEMENTS, Isotope  # type: ignore


def analyze(
    formula: str,
    /,
    *,
    maxatoms: int = 512,
    min_intensity: float = 1e-4,
    debug: bool = False,
) -> str:
    """Return analysis of chemical formula as string.

    Parameters:
        formula: Chemical formula.
        maxatoms: Number of atoms below which to calculate spectrum.
        min_intensity: Minimum intensity to include in spectrum.
        debug: If True, raise exceptions, else print error message.

    Examples:
        >>> print(analyze('C8H10N4O2', min_intensity=0.01))
        Formula: C8H10N4O2
        Empirical formula: C4H5N2O
        <BLANKLINE>
        Nominal mass: 194
        Average mass: 194.19095
        Monoisotopic mass: 194.08038 (89.883%)
        Most abundant mass: 194.08038 (89.883%)
        Mean of distribution: 194.18604
        <BLANKLINE>
        Number of atoms: 24
        <BLANKLINE>
        Elemental Composition
        <BLANKLINE>
        Element  Count  Relative mass  Fraction %
        C            8       96.08592     49.4801
        H           10       10.07941      5.1905
        N            4       56.02681     28.8514
        O            2       31.99881     16.4780
        <BLANKLINE>
        Mass Distribution
        <BLANKLINE>
        A    Relative mass  Fraction %  Intensity %
        194      194.08038   89.882781   100.000000
        195      195.08287    9.262511    10.305100
        196      196.08497    0.802196     0.892492
        197      197.08721    0.050048     0.055681

    """
    result: list[str] = []
    try:
        f = Formula(formula)
        c = f.composition()
        if f.atoms < maxatoms:
            s = f.spectrum(min_intensity=min_intensity)
        else:
            s = None

        if len(str(f)) <= 50:
            result.append(f'Formula: {f}')
        if formula != f.formula:
            result.append(f'Hill notation: {f.formula}')
        if f.formula != f.empirical:
            result.append(f'Empirical formula: {f.empirical}')

        prec = precision_digits(f.mass, 9)
        result.append(f'\nNominal mass: {f.nominal_mass}')
        result.append(f'Average mass: {f.mass:.{prec}f}')
        result.append(
            'Monoisotopic mass: '
            f'{f.isotope.mass:.{prec}f} ({f.isotope.abundance * 100:.3f}%)'
        )
        if s is not None:
            result.append(
                'Most abundant mass: '
                f'{s.peak.mass:.{prec}f} ({s.peak.fraction * 100:.3f}%)'
            )
            result.append(f'Mean of distribution: {s.mean:.{prec}f}')

        result.append(f'\nNumber of atoms: {f.atoms}')

        if len(c) > 1:
            result.extend(('\nElemental Composition\n', str(c)))

        if s is not None:
            if len(s) > 1:
                result.append(f'\nMass Distribution\n\n{s}')

    except Exception as exc:
        if debug:
            raise
        result.append(f'Error: {exc}')

    return '\n'.join(result)


class Formula:
    """Chemical formula.

    Calculate various properties from formula string, such as hill notation,
    empirical formula, mass, elemental composition, and mass distribution.

    Parameters:
        formula:
            Chemical formula. May contain only symbols of chemical elements,
            groups, isotopes, parentheses, numbers, and trailing charge.
            The formula is not validated until it is parsed on demand
            during instance attribute access or method calls.
        groups:
            Mapping of chemical group name to formula in Hill notation.
            The default is :py:attr:`GROUPS`.
        parse_groups:
            Parse chemical group names. Enabled by default.
        parse_oligos:
            Parse pure peptide and nucleotide sequences. Enabled by default.
        parse_fractions:
            Parse list of mass fractions. Enabled by default.
        parse_arithmetic:
            Parse simple arithmetic operators. Enabled by default.

    Examples:
        Elements and counts:

        >>> Formula('H2O')
        Formula('H2O')

        Isotopes:

        >>> Formula('D2O')
        Formula('[2H]2O')

        >>> Formula('[30Si]3O2')
        Formula('[30Si]3O2')

        Ion charges:

        >>> Formula('[AsO4]3-')
        Formula('[AsO4]3-')

        >>> Formula('O2-2')
        Formula('[O2]2-')

        Abbreviations of chemical groups:

        >>> Formula('EtOH')
        Formula('(C2H5)OH')

        Simple arithmetic:

        >>> Formula('(COOH)2')
        Formula('(COOH)2')

        >>> Formula('CuSO4.5H2O')
        Formula('CuSO4(H2O)5')

        Relative element weights:

        >>> Formula('O: 0.26, 30Si: 0.74')
        Formula('O2[30Si]3')

        Nucleotide sequences:

        >>> Formula('CGCGAATTCGCG')
        Formula('((C10H12N5O5P)2(C9H12N3O6P)4(C10H12N5O6P)4(C10H13N2O7P)2H2O)')

        >>> Formula('dsrna(CCUU)')
        Formula('((C10H12N5O6P)2(C9H12N3O7P)2...(C9H11N2O8P)2(H2O)2)')

        Peptide sequences:

        >>> Formula('MDRGEQGLLK')
        Formula('((C4H5NO3)(C5H7NO3)(C2H3NO)2...(C6H12N4O)H2O)')

        >>> Formula('peptide(CPK)')
        Formula('((C3H5NOS)(C6H12N2O)(C5H7NO)H2O)')

    """

    _charge: int
    _formula: str
    _formula_nocharge: str

    def __init__(
        self,
        formula: str = '',
        /,
        groups: dict[str, str] | None = None,
        *,
        parse_groups: bool = True,
        parse_oligos: bool = True,
        parse_fractions: bool = True,
        parse_arithmetic: bool = True,
    ) -> None:
        self._formula = from_string(
            formula,
            groups,
            parse_groups,
            parse_oligos,
            parse_fractions,
            parse_arithmetic,
        )
        self._formula_nocharge, self._charge = split_charge(self._formula)

    @cached_property
    def _elements(self) -> dict[str, dict[int, int]]:
        """Number of atoms and isotopes by element.

        A ``dict[symbol: dict[massnumber: count]]``, where `massnumber` is
        either an isotope mass number or zero for a natural distribution
        of isotopes.

        Raises :py:class:`FormulaError` if formula is invalid.

        >>> Formula('H')._elements
        {'H': {0: 1}}
        >>> Formula('[2H]2O')._elements
        {'O': {0: 1}, 'H': {2: 2}}

        :meta public:

        """
        formula = self._formula_nocharge
        if not formula:
            raise FormulaError('empty formula', formula, 0)

        validchars = set('([{<123456789ABCDEFGHIKLMNOPRSTUVWXYZ')

        if formula[0] not in validchars:
            raise FormulaError(
                f'unexpected character {formula[0]!r}', formula, 0
            )

        validchars |= set(']})>0abcdefghiklmnoprstuy')

        elements: dict[str, dict[int, int]] = {}
        ele = ''  # parsed element
        num = 0  # number
        level = 0  # parenthesis level
        counts = [1]  # parenthesis level multiplication
        i = len(formula)
        while i:
            i -= 1
            char = formula[i]
            if char not in validchars:
                raise FormulaError(
                    f'unexpected character {char!r}', formula, i
                )
            if char in '([{<':
                level -= 1
                if level < 0 or num != 0:
                    raise FormulaError(
                        "missing closing parenthesis ')]}>'", formula, i
                    )
            elif char in ')]}>':
                if num == 0:
                    num = 1
                level += 1
                if level > len(counts) - 1:
                    counts.append(0)
                counts[level] = num * counts[level - 1]
                num = 0
            elif char.isdigit():
                j = i
                while i and formula[i - 1].isdigit():
                    i -= 1
                num = int(formula[i : j + 1])
                if num == 0:
                    raise FormulaError('count is zero', formula, i)
            elif char.islower():
                if not formula[i - 1].isupper():
                    raise FormulaError(
                        f'unexpected character {char!r}', formula, i
                    )
                ele = char
            elif char.isupper():
                ele = char + ele
                if num == 0:
                    num = 1
                if ele not in ELEMENTS:
                    raise FormulaError(f'unknown symbol {ele!r}', formula, i)
                iso_str = ''
                j = i
                while i and formula[i - 1].isdigit():
                    i -= 1
                    iso_str = formula[i] + iso_str
                if iso_str and i and formula[i - 1] not in '([{<':
                    i = j
                    iso_str = ''
                if iso_str:
                    iso = int(iso_str)
                    if iso not in ELEMENTS[ele].isotopes:
                        raise FormulaError(
                            f"unknown isotope '{iso}{ele}'", formula, i
                        )
                else:
                    iso = 0
                number = num * counts[level]
                if ele in elements:
                    item = elements[ele]
                    if iso in item:
                        item[iso] += number
                    else:
                        item[iso] = number
                else:
                    elements[ele] = {iso: number}
                ele = ''
                num = 0

        if num != 0:
            raise FormulaError('number preceding formula', formula, 0)

        if level != 0:
            raise FormulaError(
                "missing opening parenthesis '([{<'", formula, 0
            )

        if not elements:
            raise FormulaError('invalid formula', formula, 0)

        return elements

    @cached_property
    def formula(self) -> str:
        """Formula string in Hill notation.

        >>> Formula('BrC2H5').formula
        'C2H5Br'
        >>> Formula('[(CH3)3Si2]2NNa').formula
        'C6H18NNaSi4'

        """
        return from_elements(self._elements, charge=self._charge)

    @cached_property
    def empirical(self) -> str:
        """Empirical formula in Hill notation.

        The empirical formula has the simplest whole number ratio of atoms
        of each element present in the formula.

        >>> Formula('H2O').empirical
        'H2O'
        >>> Formula('C6H12O6').empirical
        'CH2O'
        >>> Formula('[SO4]2_4-').empirical
        '[O4S]2-'

        """
        return from_elements(
            self._elements, charge=self._charge, divisor=self.gcd
        )

    @property
    def expanded(self) -> str:
        """Formula string with expanded groups, charges, and sequences.

        The input formula parsed by :py:func:`from_string`.

        >>> Formula('WQ').expanded  # peptide sequence
        '((C5H8N2O2)(C11H10N2O)H2O)'

        """
        return self._formula

    @cached_property
    def atoms(self) -> int:
        """Number of atoms.

        >>> Formula('CH3COOH').atoms
        8

        """
        return sum(sum(i.values()) for i in self._elements.values())

    @property
    def charge(self) -> int:
        """Charge number in units of elementary charge.

        >>> Formula('SO4_2-').charge
        -2

        """
        return self._charge

    @cached_property
    def gcd(self) -> int:
        """Greatest common divisor of element counts.

        >>> Formula('H2').gcd
        2
        >>> Formula('H2O').gcd
        1
        >>> Formula('C6H12O6').gcd
        6

        """
        values: list[int] = []
        for j in self._elements.values():
            values.extend(j.values())
        if abs(self._charge) > 0:
            values.append(abs(self._charge))
        return gcd(values)

    @cached_property
    def mass(self) -> float:
        """Average relative molecular mass.

        The sum of the relative atomic masses of all atoms and charges in
        the formula.
        Equals the molar mass in g/mol, that is, the mass of one mole of
        substance.

        >>> Formula('H').mass
        1.007941
        >>> Formula('H+').mass
        1.007392...
        >>> Formula('SO4_2-').mass
        96.06351...
        >>> Formula('12C').mass
        12.0
        >>> Formula('C8H10N4O2').mass
        194.1909...
        >>> Formula('C48H32AgCuO12P2Ru4').mass
        1438.404...

        """
        result = 0.0
        for symbol, massnumber_counts in self._elements.items():
            ele = ELEMENTS[symbol]
            for massnumber, count in massnumber_counts.items():
                if massnumber:
                    result += ele.isotopes[massnumber].mass * count
                else:
                    result += ele.mass * count
        return result - ELECTRON.mass * self._charge

    @property
    def monoisotopic_mass(self) -> float:
        """Mass of isotope composed of most abundant elemental isotopes.

        >>> Formula('C8H10N4O2').monoisotopic_mass
        194.08037...

        """
        return self.isotope.mass

    @property
    def nominal_mass(self) -> int:
        """Monoisotopic mass number.

        The number of protons and neutrons in the isotope composed of the most
        abundant elemental isotopes.

        >>> Formula('C8H10N4O2').nominal_mass
        194

        """
        return self.isotope.massnumber

    @property
    def mz(self) -> float:
        """Mass-to-charge ratio.

        >>> Formula('H').mz
        1.007941
        >>> Formula('H+').mz
        1.007392...
        >>> Formula('SO4_2-').mz
        48.03175...

        """
        if self._charge == 0:
            return self.mass
        return self.mass / abs(self._charge)

    @cached_property
    def isotope(self) -> Isotope:
        """Isotope composed of most abundant elemental isotopes.

        >>> Formula('C').isotope.mass
        12.0
        >>> Formula('13C').isotope.massnumber
        13
        >>> Formula('C48H32AgCuO12P2Ru4').isotope
        Isotope(mass=1439.588..., abundance=0.00205..., massnumber=1440...)

        """
        result = Isotope(-ELECTRON.mass * self._charge, 1.0, 0, self._charge)
        for symbol, massnumber_counts in self._elements.items():
            ele = ELEMENTS[symbol]
            for massnumber, count in massnumber_counts.items():
                if massnumber != 0:
                    isotope = ele.isotopes[massnumber]
                else:
                    isotope = ele.isotopes[ele.nominalmass]
                result.mass += isotope.mass * count
                result.massnumber += isotope.massnumber * count
                result.abundance *= isotope.abundance**count
        return result

    def composition(self, isotopic: bool = True) -> Composition:
        """Return elemental composition.

        Parameters:
            isotopic:
                List isotopes separately as opposed to part of an element.

        Examples:
            >>> print(Formula('[12C]C').composition())
            Element  Count  Relative mass  Fraction %
            C            1      12.010740     50.0224
            12C          1      12.000000     49.9776

            >>> print(Formula('[12C]C').composition(False))
            Element  Count  Relative mass  Fraction %
            C            2      24.010740    100.0000

        """
        elements = self._elements
        result = []
        if isotopic:
            for symbol in hill_sorted(elements):
                ele = ELEMENTS[symbol]
                iso = elements[symbol]
                for massnumber in sorted(iso):
                    count = iso[massnumber]
                    if massnumber:
                        mass = ele.isotopes[massnumber].mass * count
                        symbol = f'{massnumber}{symbol}'
                    else:
                        mass = ele.mass * count
                    result.append((symbol, count, mass, mass / self.mass))
        else:
            for symbol in hill_sorted(elements):
                ele = ELEMENTS[symbol]
                mass = 0.0
                counter = 0
                for massnumber, count in elements[symbol].items():
                    counter += count
                    if massnumber:
                        mass += ele.isotopes[massnumber].mass * count
                    else:
                        mass += ele.mass * count
                result.append((symbol, counter, mass, mass / self.mass))
        if self._charge != 0:
            mass = -ELECTRON.mass * self._charge
            result.append(('e-', -self._charge, mass, mass / self.mass))
        return Composition(result)

    def spectrum(
        self,
        *,
        min_fraction: float = 1e-16,
        min_intensity: float | None = None,
    ) -> Spectrum:
        """Return low resolution mass spectrum.

        Calculated by combining the mass numbers of the elemental isotopes.

        Parameters:
            min_fraction: Minimum of fraction to return.
            min_intensity: Minimum intensity to return.

        Returns:
            Mapping of massnumber to mass, fraction, and intensity.

        Examples:
            >>> print(Formula('D').spectrum())
            A  Relative mass  Fraction %  Intensity %
            2      2.0141018  100.000000   100.000000

            >>> print(Formula('H').spectrum())
            A  Relative mass  Fraction %  Intensity %
            1      1.0078250   99.988500   100.000000
            2      2.0141018    0.011500     0.011501

            >>> print(Formula('H+').spectrum())
            A  Relative mass  Fraction %  Intensity %
            1      1.0072765   99.988500   100.000000
            2      2.0135532    0.011500     0.011501

            >>> print(Formula('D2').spectrum())
            A  Relative mass  Fraction %  Intensity %
            4      4.0282036  100.000000   100.000000

            >>> print(Formula('DH').spectrum())
            A  Relative mass  Fraction %  Intensity %
            3      3.0219268   99.988500   100.000000
            4      4.0282036    0.011500     0.011501

            >>> print(Formula('H2').spectrum())
            A  Relative mass  Fraction %  Intensity %
            2      2.0156501   99.977001   100.000000
            3      3.0219268    0.022997     0.023003
            4      4.0282036    0.000001     0.000001

            >>> print(Formula('DHO').spectrum())
            A   Relative mass  Fraction %  Intensity %
            19      19.016841   99.745528   100.000000
            20      20.021536    0.049468     0.049594
            21      21.021087    0.204981     0.205504
            22      22.027363    0.000024     0.000024

            >>> print(Formula('SO4_2-').spectrum(min_intensity=0.1))
            A   Relative mass  Fraction %  Intensity %  m/z
            96      95.952827   94.070057   100.000000  47.976413
            97      96.952996    0.886071     0.941927  48.476498
            98      97.949936    4.983307     5.297443  48.974968

        """
        spectrum: dict[int, list[float]] = {0: [0.0, 1.0, 0.0]}
        elements: dict[str, dict[int, int]] = self._elements

        for symbol, massnumber_counts in elements.items():
            ele = ELEMENTS[symbol]
            for massnumber, count in massnumber_counts.items():
                if massnumber:
                    # specific isotope
                    iso = ele.isotopes[massnumber]
                    for key in reversed(sorted(spectrum)):
                        t = spectrum[key]
                        del spectrum[key]
                        if t[1] < min_fraction:
                            continue
                        f = t[1]
                        m = t[0] + iso.mass * count
                        k = key + iso.massnumber * count
                        if k in spectrum:
                            s = spectrum[k]
                            s[0] += (s[1] * s[0] + f * m) / (s[1] + f)
                            s[1] += f
                        else:
                            spectrum[k] = [m, f]
                else:
                    # mixture of isotopes
                    isotopes = ele.isotopes.values()
                    for _ in range(count):
                        for key in reversed(sorted(spectrum)):
                            t = spectrum[key]
                            del spectrum[key]
                            if t[1] < min_fraction:
                                continue
                            for iso in isotopes:
                                f = t[1] * iso.abundance
                                m = t[0] + iso.mass
                                k = key + iso.massnumber
                                if k in spectrum:
                                    s = spectrum[k]
                                    s[0] = (s[1] * s[0] + f * m) / (s[1] + f)
                                    s[1] += f
                                else:
                                    spectrum[k] = [m, f]

        # filter low intensities
        if min_intensity is not None:
            norm = 100 / max(v[1] for v in spectrum.values())
            for massnumber, value in spectrum.copy().items():
                if value[1] * norm < min_intensity:
                    del spectrum[massnumber]

        if self._charge != 0:
            # correct for electron mass
            mc = ELECTRON.mass * self._charge
            for value in spectrum.values():
                value[0] -= mc

        return Spectrum(spectrum, self._charge)

    def __mul__(self, number: int, /) -> Formula:
        """Return formula repeated number times.

        >>> Formula('HO-') * 2
        Formula('[(HO)2]2-')

        """
        if not isinstance(number, int) or number < 1:
            raise TypeError('can only multiply with positive number')
        return Formula(
            join_charge(
                f'({self._formula_nocharge}){number}', self._charge * number
            )
        )

    def __rmul__(self, number: int, /) -> Formula:
        """Return formula repeated number times.

        >>> 2 * Formula('H2O')
        Formula('(H2O)2')

        """
        return self.__mul__(number)

    def __add__(self, other: Formula, /) -> Formula:
        """Add this and other formula.

        >>> Formula('H2O') + Formula('HO-')
        Formula('[(H2O)(HO)]-')

        """
        if not isinstance(other, Formula):
            raise TypeError('can only add Formula instance')
        return Formula(
            join_charge(
                f'({self._formula_nocharge})({other._formula_nocharge})',
                self._charge + other.charge,
            )
        )

    def __sub__(self, other: Formula, /) -> Formula:
        """Subtract elements of other formula.

        >>> Formula('H2O') - Formula('O')
        Formula('H2')

        """
        if not isinstance(other, Formula):
            raise TypeError('can only subtract Formula instance')
        _elements = copy.deepcopy(self._elements)
        for symbol, isotopes in other._elements.items():
            if symbol not in _elements:
                raise ValueError(f'element {symbol} not in {self}')
            element = _elements[symbol]
            for massnumber, count in isotopes.items():
                if massnumber not in element:
                    raise ValueError(
                        f'element {massnumber}{symbol} not in {self}'
                    )
                element[massnumber] -= count
                if element[massnumber] < 0:
                    raise ValueError(
                        f'negative number of element {massnumber}{symbol}'
                    )
                if element[massnumber] == 0:
                    del element[massnumber]
                if not element:
                    del _elements[symbol]
        return Formula(
            from_elements(_elements, charge=self._charge - other.charge)
        )

    def __str__(self) -> str:
        return self._formula

    def __repr__(self) -> str:
        return f'Formula({self._formula!r})'


@dataclass
class CompositionItem:
    """Item of elemental composition."""

    symbol: str
    """Chemical symbol of element or electron."""

    count: int
    """Number of elements or electrons in composition."""

    mass: float
    """Relative mass."""

    fraction: float
    """Mass fraction."""

    def astuple(self) -> tuple[str, int, float, float]:
        """Return (symbol, count, mass, fraction)."""
        return self.symbol, self.count, self.mass, self.fraction


class Composition:
    """Elemental composition.

    The basic interface is an ordered ``dict[symbol, CompositionItem]]``.

    Parameters:
        items:
            Symbol, count, mass, and fraction for all elements in formula.
            Symbols must not be repeated.

    Examples:
        >>> c = Composition((('2H', 2, 4.028, 0.201), ('O', 1, 15.999, 0.799)))
        >>> c
        <Composition([('2H', 2, 4.028, 0.201), ...])>
        >>> print(c)
        Element  Count  Relative mass  Fraction %
        2H           2       4.028000     20.1000
        O            1      15.999000     79.9000
        >>> c['O']
        CompositionItem(symbol='O', count=1, mass=15.999, fraction=0.799)

    """

    _items: dict[str, CompositionItem]

    def __init__(self, items: Sequence[tuple[str, int, float, float]]) -> None:
        self._items = {
            e[0]: CompositionItem(e[0], e[1], e[2], e[3]) for e in items
        }

    @cached_property
    def total(self) -> CompositionItem:
        """Sums of counts (excluding electrons), masses, and fractions.

        >>> Formula('2HO').composition().total
        CompositionItem(symbol='Total', count=2, mass=18.0135..., fraction=1.0)

        """
        result = CompositionItem('Total', 0, 0.0, 0.0)
        for item in self._items.values():
            if item.symbol != 'e-':
                result.count += item.count
            result.mass += item.mass
            result.fraction += item.fraction
        return result

    def dataframe(self) -> pandas.DataFrame:
        """Return composition as pandas DataFrame.

        >>> Formula('2HO').composition().dataframe()
                 Count  Relative mass  Fraction
        Element...
        2H           1       2.014102  0.111811
        O            1      15.999405  0.888189

        """
        from pandas import DataFrame

        data = {
            'Element': [i.symbol for i in self._items.values()],
            'Count': [i.count for i in self._items.values()],
            'Relative mass': [i.mass for i in self._items.values()],
            'Fraction': [i.fraction for i in self._items.values()],
        }
        return DataFrame(data).set_index('Element')

    def asdict(self) -> dict[str, tuple[int, float, float]]:
        """Return dict[symbol, tuple[symbol, count, mass, fraction]].

        >>> Formula('2HO').composition().asdict()
        {'2H': (1, 2.0141..., 0.1118...), 'O': (1, 15.9994..., 0.8881...)}

        """
        return {key: value.astuple()[1:] for key, value in self._items.items()}

    def astuple(self) -> tuple[tuple[str, int, float, float], ...]:
        """Return tuple[tuple[symbol, count, mass, fraction], ...].

        >>> Formula('2HO').composition().astuple()
        (('2H', 1, 2.0141..., 0.1118...), ('O', 1, 15.9994..., 0.8881...))

        """
        return tuple(value.astuple() for value in self._items.values())

    def keys(self) -> Iterator[str]:
        return iter(self._items.keys())

    def values(self) -> Iterator[CompositionItem]:
        return iter(self._items.values())

    def items(self) -> Iterator[tuple[str, CompositionItem]]:
        return iter(self._items.items())

    def __getitem__(self, key: str) -> CompositionItem:
        return self._items[key]

    def __len__(self) -> int:
        return len(self._items)

    def __iter__(self) -> Iterator[str]:
        return iter(self._items)

    def __repr__(self) -> str:
        if len(self._items) > 1:
            return f'<Composition([{self.astuple()[0]}, ...])>'
        return f'Composition({list(self.astuple())})'

    def __str__(self) -> str:
        if len(self) == 0:
            return ''
        precision = precision_digits(self.total.mass, 9)
        result = ['Element  Count  Relative mass  Fraction %']
        for item in self._items.values():
            result.append(
                f'{item.symbol:<7s}'
                f'  {item.count:5}'
                f'  {item.mass:13.{precision}f}'
                f'  {item.fraction * 100:10.4f}'
            )
        return '\n'.join(result)


@dataclass
class SpectrumEntry:
    """Entry of mass distribution spectrum."""

    massnumber: int
    """Mass number, bin of spectrum."""

    mass: float
    """Relative mass."""

    fraction: float
    """Mass fraction or abundance."""

    intensity: float
    """Fraction normalized by maximum fraction in %."""

    mz: float
    """Mass-to-charge ratio."""

    def astuple(self) -> tuple[int, float, float, float, float]:
        """Return attributes (massnumber, mass, fraction, intensity, mz)."""
        return (
            self.massnumber,
            self.mass,
            self.fraction,
            self.intensity,
            self.mz,
        )


class Spectrum:
    """Mass distribution.

    The basic interface is a sorted ``dict[massnumber, SpectrumEntry]``.

    Parameters:
        spectrum: Mapping of mass number to relative mass and fraction.
        charge: Charge number.

    Examples:
        >>> s = Spectrum({1: [1.078, 0.9999], 2: [2.014, 0.0001]})
        >>> s
        <Spectrum({1: [1, 1.078, 0.9999, 100.0, 1.078], ...})>
        >>> print(s)
        A  Relative mass  Fraction %  Intensity %
        1      1.0780000   99.990000   100.000000
        2      2.0140000    0.010000     0.010001
        >>> s[2]
        SpectrumEntry(massnumber=2, mass=2.014, fraction=0.0001, ..., mz=2.014)

    """

    _spectrum: dict[int, SpectrumEntry]
    _charge: int

    def __init__(
        self, spectrum: dict[int, list[float]], /, charge: int = 0
    ) -> None:
        mz = 1.0 / (1 if charge == 0 else abs(charge))
        intensity = 100 / max(v[1] for v in spectrum.values())
        self._spectrum = {
            massnumber: SpectrumEntry(
                massnumber=massnumber,
                mass=entry[0],
                fraction=entry[1],
                intensity=entry[1] * intensity,
                mz=entry[0] * mz,
            )
            for massnumber, entry in sorted(spectrum.items())
        }
        self._charge = charge

    @cached_property
    def peak(self) -> SpectrumEntry:
        """Most abundant entry.

        >>> Formula('C8H10N4O2').spectrum().peak.mass
        194.0803...

        """
        return max(self._spectrum.values(), key=lambda x: x.fraction)

    @cached_property
    def mean(self) -> float:
        """Mean of all masses.

        >>> Formula('C8H10N4O2').spectrum().mean
        194.1909...

        """
        return sum((v.mass * v.fraction) for v in self._spectrum.values())

    @cached_property
    def range(self) -> tuple[int, int]:
        """Smallest and largest massnumbers.

        >>> Formula('C8H10N4O2').spectrum().range
        (194, 205)

        """
        return min(self._spectrum.keys()), max(self._spectrum.keys())

    def dataframe(self) -> pandas.DataFrame:
        """Return composition as pandas DataFrame.

        >>> Formula('SO4_2-').spectrum(min_intensity=0.1).dataframe()
                     Relative mass  Fraction  Intensity %        m/z
        Mass number...
        96               95.952827  0.940701   100.000000  47.976413
        97               96.952996  0.008861     0.941927  48.476498
        98               97.949936  0.049833     5.297443  48.974968

        """
        from pandas import DataFrame

        data = {
            'Mass number': [i.massnumber for i in self._spectrum.values()],
            'Relative mass': [i.mass for i in self._spectrum.values()],
            'Fraction': [i.fraction for i in self._spectrum.values()],
            'Intensity %': [i.intensity for i in self._spectrum.values()],
            'm/z': [i.mz for i in self._spectrum.values()],
        }
        return DataFrame(data).set_index('Mass number')

    def asdict(self) -> dict[int, tuple[float, float, float, float]]:
        """Return dict[massnumber, tuple[mass, fraction, intensity, mz]].

        >>> Formula('DH').spectrum().asdict()
        {3: (3.0..., 0.9..., 100.0, 3...), 4: (4.0..., 0.0..., 0.0..., 4...)}

        """
        return {
            key: value.astuple()[1:] for key, value in self._spectrum.items()
        }

    def astuple(self) -> tuple[tuple[int, float, float, float, float], ...]:
        """Return tuple[tuple[massnumber, mass, fraction, intensity, mz], ...].

        >>> Formula('DH').spectrum().astuple()
        ((3, 3.0..., 0.9..., 100.0, 3...), (4, 4.0..., 0.0..., 0.0..., 4...))

        """
        return tuple(value.astuple() for value in self._spectrum.values())

    def keys(self) -> Iterator[int]:
        return iter(self._spectrum.keys())

    def values(self) -> Iterator[SpectrumEntry]:
        return iter(self._spectrum.values())

    def items(self) -> Iterator[tuple[int, SpectrumEntry]]:
        return iter(self._spectrum.items())

    def __getitem__(self, key: int) -> SpectrumEntry:
        return self._spectrum[key]

    def __len__(self) -> int:
        return len(self._spectrum)

    def __iter__(self) -> Iterator[int]:
        return iter(self._spectrum)

    def __repr__(self) -> str:
        item = list(list(self._spectrum.values())[0].astuple())
        return '<Spectrum({' + f'{item[0]!r}: {item!r}, ...' + '})>'

    def __str__(self) -> str:
        if not self._spectrum:
            return ''
        a = len(str(self.range[-1]))
        mz = '  m/z' if abs(self._charge) > 1 else ''
        result = [f'A{" " * a} Relative mass  Fraction %  Intensity %{mz}']
        precision = precision_digits(self.peak.mass, 9)
        for entry in self.values():
            mz = f'  {entry.mz:<13.{precision}f}' if mz else ''
            result.append(
                f'{entry.massnumber:<{a}d}'
                f'  {entry.mass:>13.{precision}f}'
                f' {entry.fraction*100:11.6f}'
                f' {entry.intensity:12.6f}'
                f'{mz}'.strip()
            )
        return '\n'.join(result)


class FormulaError(Exception):
    """Error in chemical formula.

    Parameters:
        message: Error message.
        formula: Chemical formula in which error occurred.
        position: Position in formula where error occurred.

    Examples:
        >>> Formula('abc').formula
        Traceback (most recent call last):
         ...
        molmass.molmass.FormulaError: unexpected character 'a'
        abc
        ^

        >>> Formula('(H2O)2-H2O').formula
        Traceback (most recent call last):
         ...
        molmass.molmass.FormulaError: subtraction not allowed
        (H2O)2-H2O
        ......^

        >>> Formula('[11C]').formula
        Traceback (most recent call last):
         ...
        molmass.molmass.FormulaError: unknown isotope '11C'
        [11C]
        .^

    """

    message: str
    """Error message."""

    formula: str
    """Formula in which error occurred."""

    position: int
    """Position in formula where error occurred."""

    def __init__(
        self, message: str, formula: str = '', position: int = -1
    ) -> None:
        self.position = position
        self.message = message
        self.formula = formula
        Exception.__init__(self, message, formula, position)

    def __str__(self) -> str:
        if self.position < 0:
            return str(self.message)
        return f"{self.message}\n{self.formula}\n{'.' * self.position}^"


def from_string(
    formula: str,
    /,
    groups: dict[str, str] | None = None,
    parse_groups: bool = True,
    parse_oligos: bool = True,
    parse_fractions: bool = True,
    parse_arithmetic: bool = True,
) -> str:
    r"""Return formula string from user input string.

    Parameters:
        formula:
            Chemical formula.
            Supports simple, non-nested arithmetic (\+ and \*), abbreviations
            of common chemical groups, peptides, oligos, and mass fractions.
        groups:
            Mapping of chemical group name to formula in Hill notation.
            The default is :py:attr:`GROUPS`.
        parse_groups:
            Parse chemical group names.
        parse_oligos:
            Parse pure peptide and nucleotide sequences.
        parse_fractions:
            Parse list of mass fractions.
        parse_arithmetic:
            Parse simple arithmetic operators.

    Returns:
        Chemical formula composed only of element and isotope symbols,
        parentheses, numbers, and trailing ion charges.

    Raises:
        FormulaError: Invalid formula.

    Examples:
        >>> from_string('Valohp')
        '(C5H8NO2)'
        >>> from_string('Valohp', parse_groups=False)
        'Valohp'
        >>> from_string('HLeu2OH')
        'H(C6H11NO)2OH'
        >>> from_string('D2O')
        '[2H]2O'
        >>> from_string('O: 0.26, 30Si: 0.74')
        'O2[30Si]3'
        >>> from_string('O: 0.26, 30Si: 0.74', parse_fractions=False)
        'O:0(,30Si:0)26()74'
        >>> from_string('PhNH2.HCl')
        '(C6H5)NH2HCl'
        >>> from_string('PhNH2.HCl', parse_arithmetic=False)
        '(C6H5)NH2.HCl'
        >>> from_string('CuSO4.5H2O')
        'CuSO4(H2O)5'
        >>> from_string('CuSO4+5*H2O')
        'CuSO4(H2O)5'
        >>> from_string('5*H2O')
        '(H2O)5'
        >>> from_string('O2+12C')  # not [12C]O2
        'O2(C)12'
        >>> from_string('ssdna(AC)')
        '((C10H12N5O5P)(C9H12N3O6P)H2O)'
        >>> from_string('peptide(GG)')
        '((C2H3NO)2H2O)'
        >>> from_string('WQ')
        '((C5H8N2O2)(C11H10N2O)H2O)'
        >>> from_string('WQ', parse_oligos=False)
        'WQ'

    """
    try:
        formula = formula.strip().replace(' ', '')
    except AttributeError as exc:
        raise ValueError('formula must be a string') from exc

    if parse_groups:
        if groups is None:
            groups = GROUPS
        if groups:
            for grp in reversed(sorted(groups)):
                formula = formula.replace(grp, f'({groups[grp]})')

    if parse_fractions and ':' in formula and ',' in formula:
        fractions: dict[str, float] = {}
        try:
            for item in formula.split(','):
                items = item.split(':')
                fractions[items[0].strip()] = float(items[1].strip())
        except Exception as exc:
            raise FormulaError(
                'invalid list of mass fractions', formula
            ) from exc
        return from_fractions(fractions)

    if parse_oligos and len(formula) > 1:
        fset = set(formula)
        if fset <= set('ATCG') and fset & set('ATG'):
            return from_oligo(formula, 'ssdna')
        if fset <= set('AUCG') and fset & set('AG'):
            return from_oligo(formula, 'ssrna')
        if fset <= set(AMINOACIDS.keys()) and fset & set('AEGMLQRT'):
            return from_peptide(formula)
        for dtype, func in PREPROCESSORS.items():
            for match in re.findall(dtype + r'\((.*?)\)', formula):
                formula = formula.replace(f'{dtype}({match})', func(match))

    # Deuterium
    formula = re.sub('(D)(?![a-z])', '[2H]', formula)

    # charge
    formula, charge = split_charge(formula)

    if parse_arithmetic:
        formula = formula.replace('.', '+')
        if '+' in formula or '*' in formula:
            # TODO: document O2+12C is C12O2, not [12C]O2
            for match in re.findall(
                r'(?:\+|^)((\d+)\*?(.*?))(?:(?=\+)|$)', formula
            ):
                formula = formula.replace(match[0], f'({match[2]}){match[1]}')
            m = re.search(r'(\+[a-z])', formula)
            if m:
                raise FormulaError(
                    'invalid symbol', formula, formula.index(m[0]) + 1
                )
            formula = formula.replace('+', '')
        if '-' in formula:
            raise FormulaError(
                'subtraction not allowed', formula, formula.index('-')
            )

    if charge != 0:
        formula = f'[{formula}]{format_charge(charge)}'
    return formula


def from_elements(
    elements: dict[str, dict[int, int]],
    /,
    *,
    divisor: int = 1,
    charge: int = 0,
    fmt: tuple[str, str, str, str] | None = None,
) -> str:
    """Return formula string in Hill notation from elements dict.

    Parameters:
        elements:
            Number of atoms and isotopes by element.
            See :py:meth:`Formula._elements`.
        divisor:
            Number by which to divide element counts and charge.
        charge:
            Charge number.
        fmt:
            Format strings.

    Examples:
        >>> from_elements({'C': {0: 4, 12: 2}}, divisor=2)
        'C2[12C]'

        >>> from_elements(
        ...     {'C': {0: 4, 12: 2}},
        ...     divisor=2,
        ...     fmt=(
        ...         '{}',
        ...         '{}<sub>{}</sub>',
        ...         '<sup>{}</sup>{}',
        ...         '<sup>{}</sup>{}<sub>{}</sub>',
        ...     ),
        ... )
        'C<sub>2</sub><sup>12</sup>C'

    """
    if fmt is None:
        fmt = ('{}', '{}{}', '[{}{}]', '[{}{}]{}')
    formula = []
    for symbol in hill_sorted(elements):
        isotopes = elements[symbol]
        for massnumber in sorted(isotopes):
            count = isotopes[massnumber] // divisor
            if massnumber:
                if count == 1:
                    formula.append(fmt[2].format(massnumber, symbol))
                else:
                    formula.append(fmt[3].format(massnumber, symbol, count))
            else:
                if count == 1:
                    formula.append(fmt[0].format(symbol))
                else:
                    formula.append(fmt[1].format(symbol, count))
    return join_charge(''.join(formula), charge // divisor)


def from_fractions(
    fractions: dict[str, float],
    /,
    *,
    maxcount: int = 10,
    precision: float = 1e-4,
) -> str:
    """Return formula string from elemental mass fractions.

    Parameters:
        fractions: Mapping of element symbols to abundances.
        maxcount: Maximum count of single element in formula.
        precision: Threshold for finding smallest acceptable element count.

    Examples:
        >>> from_fractions({'H': 0.112, 'O': 0.888})
        'H2O'
        >>> from_fractions({'D': 0.2, 'O': 0.8})
        'O[2H]2'
        >>> from_fractions({'H': 8.97, 'C': 59.39, 'O': 31.64})
        'C5H9O2'
        >>> from_fractions({'O': 0.26, '30Si': 0.74})
        'O2[30Si]3'

    """
    if not fractions:
        return ''
    # divide normalized fractions by element/isotope mass
    numbers = {}
    sumfractions = sum(fractions.values())
    for symbol, fraction in fractions.items():
        if symbol == 'D':  # Deuterium
            symbol = '2H'
        if symbol[0].isupper():
            try:
                mass = ELEMENTS[symbol].mass
            except KeyError as exc:
                raise FormulaError(f'unknown element {symbol!r}') from exc
        else:
            if symbol.startswith('['):
                symbol = symbol[1:-1]
            i = 0
            while symbol[i].isdigit():
                i += 1
            massnum = int(symbol[:i])
            symbol = symbol[i:]
            try:
                mass = ELEMENTS[symbol].isotopes[massnum].mass
            except KeyError as exc:
                raise FormulaError(
                    f"unknown isotope '[{massnum}{symbol}]'"
                ) from exc
            symbol = f'[{massnum}{symbol}]'
        numbers[symbol] = fraction / (sumfractions * mass)

    # divide numbers by smallest number
    smallest = min(numbers.values())
    for symbol in numbers:
        numbers[symbol] /= smallest

    # find smallest factor that turns all numbers into integers
    precision *= len(numbers)
    best = 1e6
    factor = 1
    for i in range(1, maxcount):
        x = sum(abs((i * n) - round(i * n)) for n in numbers.values())
        if x < best:
            best = x
            factor = i
            if best < i * precision:
                break

    formula = []
    for symbol, number in sorted(numbers.items()):
        count = int(round(factor * number))
        if count > 1:
            formula.append(f'{symbol}{count}')
        else:
            formula.append(symbol)
    return ''.join(formula)


def from_sequence(sequence: str, groups: dict[str, str], /) -> str:
    """Return sequence as chemical formula.

    Parameters:
        sequence:
            DNA, RNA, or peptide sequence.
        groups:
            Mapping of sequence item to chemical formula in Hill notation.
            One of :py:attr:`DEOXYNUCLEOTIDES`, :py:attr:`NUCLEOTIDES`, or
            :py:attr:`AMINOACIDS`.

    Examples:
        >>> from_sequence('A', {'A': 'B'})
        '(B)'
        >>> from_sequence('AA', {'A': 'B'})
        '(B)2'

    """
    sequence, charge = split_charge(sequence)
    counts = {key: 0 for key in groups}
    for item in sequence:
        counts[item] += 1
    formula = []
    for key in sorted(groups):
        num = counts[key]
        if num == 1:
            formula.append(f'({groups[key]})')
        elif num:
            formula.append(f'({groups[key]}){num}')
    return join_charge(''.join(formula), charge)


def from_peptide(sequence: str, /) -> str:
    """Return chemical formula for polymer of unmodified amino acids.

    Parameters:
        sequence: Perptide sequence.

    Examples:
        >>> from_peptide('GG')
        '((C2H3NO)2H2O)'

        >>> f = Formula(from_peptide('GPAVL IMCFY WHKRQ NEDST_2+'))
        >>> print(f.formula, f.atoms, f.mass)
        [C107H159N29O30S2]2+ 327 2395.7168...

    """
    sequence = sequence.replace(' ', '')
    sequence, charge = split_charge(sequence)
    formula = f'({from_sequence(sequence, AMINOACIDS)}H2O)'
    return join_charge(formula, charge)


def from_oligo(sequence: str, dtype: str = 'ssdna', /) -> str:
    """Return chemical formula for polymer of unmodified (deoxy)nucleotides.

    Each strand includes a 5' monophosphate.

    Parameters:
        sequence:
            DNA or RNA sequence.
        dtype:
            Nucleic acid sequence type.
            One of 'ssdna', 'dsdna', 'ssrna', or 'dsrna'.

    Examples:
        >>> from_oligo('AC', 'ssdna')
        '((C10H12N5O5P)(C9H12N3O6P)H2O)'

        >>> from_oligo('AU', 'dsrna')
        '((C10H12N5O6P)2(C9H11N2O8P)2(H2O)2)'

        >>> f = Formula(from_oligo('ATC G', 'dsdna'))
        >>> print(f.formula, f.atoms, f.mass)
        C78H102N30O50P8 268 2507.60913...

        >>> f = Formula(from_oligo('AUC G_2+', 'ssrna'))
        >>> print(f.formula, f.atoms, f.mass)
        [C38H49N15O29P4]2+ 135 1303.7744...

    """
    sequence, charge = split_charge(sequence)
    sequence = sequence.replace(' ', '')
    dtype_str = dtype.lower()
    if 'rna' in dtype:
        items = NUCLEOTIDES
        complements = NUCLEOTIDE_COMPLEMENTS
    else:
        items = DEOXYNUCLEOTIDES
        complements = DEOXYNUCLEOTIDE_COMPLEMENTS

    if dtype_str.startswith('ds'):
        t = ''.join(complements[i] for i in sequence)
        formula = from_sequence(sequence + t, items)
        formula = f'({formula}(H2O)2)'
    else:
        formula = from_sequence(sequence, items)
        formula = f'({formula}H2O)'
    return join_charge(formula, charge)


def hill_sorted(symbols: Iterable[str], /) -> Iterator[str]:
    """Return iterator over element symbols in order of Hill notation.

    Parameters:
        symbols: Element symbols.

    Yields:
        Element symbols in order of Hill notation.

    Examples:
        >>> tuple(hill_sorted('HCO'))
        ('C', 'H', 'O')

    """
    symbols_set = set(symbols)
    if 'C' in symbols:
        symbols_set.remove('C')
        yield 'C'
        if 'H' in symbols_set:
            symbols_set.remove('H')
            yield 'H'
    yield from sorted(symbols_set)


def gcd(numbers: Iterable[int], /) -> int:
    """Return greatest common divisor of integer numbers.

    Using Euclid's algorithm.

    Parameters:
        numbers: Integer numbers.

    Examples:
        >>> gcd([4])
        4
        >>> gcd([3, 6])
        3
        >>> gcd([6, 7])
        1

    """

    def _gcd(a: int, b: int, /) -> int:
        # return greatest common divisor of two integer numbers
        while b:
            a, b = b, a % b
        return a

    return reduce(_gcd, set(numbers))


def precision_digits(f: float, width: int, /) -> int:
    """Return number of digits after decimal point to print f in width chars.

    Parameters:
        f: Floating point number to print in `width` characters.
        width: Maximum length of printed number.

    Examples:
        >>> precision_digits(-0.12345678, 5)
        2
        >>> precision_digits(1.23456789, 5)
        3
        >>> precision_digits(12.3456789, 5)
        2
        >>> precision_digits(12345.6789, 5)
        1

    """
    precision = math.log(abs(f), 10)
    precision = max(precision, 0)
    precision = width - int(math.floor(precision))
    precision -= 3 if f < 0 else 2  # sign and decimal point
    precision = max(precision, 1)
    return precision


def split_charge(formula: str, /) -> tuple[str, int]:
    """Return formula stripped from charge, and charge.

    Parameters:
        formula: Chemical formula with appended charge.

    Returns:
        Chemical formula without charge, and charge number.

    Examples:
        >>> split_charge('Formula')
        ('Formula', 0)
        >>> split_charge('Formula+')
        ('Formula', 1)
        >>> split_charge('Formula+1')
        ('Formula', 1)
        >>> split_charge('Formula++')
        ('Formula', 2)
        >>> split_charge('[Formula]2+')
        ('Formula', 2)
        >>> split_charge('Formula+2')
        ('Formula', 2)
        >>> split_charge('[[Formula]]2-')
        ('[Formula]', -2)
        >>> split_charge('[Formula]-2')
        ('Formula', -2)
        >>> split_charge('[Formula]_2-')
        ('Formula', -2)
        >>> split_charge('Formula_2-')
        ('Formula', -2)
        >>> split_charge('Formula-2')
        ('Formula', -2)
        >>> split_charge('Formula_+')
        ('Formula', 1)
        >>> split_charge('Formula+-')
        ('Formula', 0)
        >>> split_charge('Formul+a+')
        ('Formul+a', 1)

    """
    charge = 0
    m = re.search(r'([\]_])([0-9]{1,})([+-]{1,})$', formula)
    if m:
        m_delim, m_count, m_sign = m.groups()
        if m_count == '':
            charge = int(f'{m_sign}1')
        else:
            charge = int(f'{m_sign}{m_count}')
    else:
        m = re.search(r'([+-]{1})([0-9]{1,})$', formula)
        if m:
            m_sign, m_count = m.groups()
            m_delim = f'{m_sign}{m_count}'
            charge = int(m_delim)
        else:
            m = re.search(r'([\]_]?)([+-]{1,})$', formula)
            charge = 0
            if m:
                m_delim, m_sign = m.groups()
                for char in m_sign:
                    if char == '+':
                        charge += 1
                    elif char == '-':
                        charge -= 1
    if m:
        if m_delim == '_':
            formula = formula.rsplit('_', 1)[0]
        elif m_delim == '':
            formula = formula.strip(m_sign)
        elif m_delim == ']':
            formula = formula.rsplit(']', 1)[0] + ']'
        else:
            formula = formula.rsplit(m_delim, 1)[0]
        if formula.startswith('[') and formula.endswith(']'):
            formula = formula[1:-1]
    return formula, charge


def join_charge(formula: str, charge: int, separator: str = '', /) -> str:
    """Return formula with charge appended.

    Parameters:
        formula: Chemical formula without charge.
        charge: Charge number.
        separator: Character separating formula from charge. One of '' or '_'.

    Returns:
        Formula with charge != 0 appended.
        If `separator` is empty, the formula is wrapped in square brackets.

    Examples:
        >>> join_charge('Formula', 0)
        'Formula'
        >>> join_charge('Formula', 1)
        '[Formula]+'
        >>> join_charge('Formula', 2)
        '[Formula]2+'
        >>> join_charge('Formula', -2, '_')
        'Formula_2-'

    """
    if charge != 0:
        if separator:
            formula = f'{formula}{separator}{format_charge(charge, "")}'
        else:
            formula = f'[{formula}]{format_charge(charge, "")}'
    return formula


def format_charge(charge: int, prefix: str = '', /) -> str:
    """Return string representation of charge.

    Parameters:
        charge: Charge number.
        prefix: Character placed in front of charge in case ``abs(charge)>1``.

    Examples:
        >>> format_charge(-2, '_')
        '_2-'
        >>> format_charge(1, '_')
        '+'

    """
    if charge == 0:
        return '0'
    if prefix and abs(charge) <= 1:
        prefix = ''
    sign = '+' if charge > 0 else '-'
    count = f'{abs(charge)}' if abs(charge) > 1 else ''
    return f'{prefix}{count}{sign}'


# Common chemical groups
GROUPS: dict[str, str] = {
    'Abu': 'C4H7NO',
    'Acet': 'C2H3O',
    'Acm': 'C3H6NO',
    'Adao': 'C10H15O',
    'Ade': 'C5H5N5',
    'Aib': 'C4H7NO',
    'Ala': 'C3H5NO',
    'Arg': 'C6H12N4O',
    'Argp': 'C6H11N4O',
    'Asn': 'C4H6N2O2',
    'Asnp': 'C4H5N2O2',
    'Asp': 'C4H5NO3',
    'Aspp': 'C4H4NO3',
    'Asu': 'C8H13NO3',
    'Asup': 'C8H12NO3',
    'Boc': 'C5H9O2',
    'Bom': 'C8H9O',
    'Bpy': 'C10H8N2',  # Bipyridine
    'Brz': 'C8H6BrO2',
    'Bu': 'C4H9',
    'Bum': 'C5H11O',
    'Bz': 'C7H5O',
    'Bzl': 'C7H7',
    'Bzlo': 'C7H7O',
    'Cha': 'C9H15NO',
    'Chxo': 'C6H11O',
    'Cit': 'C6H11N3O2',
    'Citp': 'C6H10N3O2',
    'Clz': 'C8H6ClO2',
    'Cp': 'C5H5',
    'Cy': 'C6H11',
    'Cys': 'C3H5NOS',
    'Cysp': 'C3H4NOS',
    'Cyt': 'C4H5N3O',
    'Dde': 'C10H13O2',
    'Dnp': 'C6H3N2O4',
    'Et': 'C2H5',
    'Fmoc': 'C15H11O2',
    'For': 'CHO',
    'Gln': 'C5H8N2O2',
    'Glnp': 'C5H7N2O2',
    'Glp': 'C5H5NO2',
    'Glu': 'C5H7NO3',
    'Glup': 'C5H6NO3',
    'Gly': 'C2H3NO',
    'Gua': 'C5H5N5O',
    'Hci': 'C7H13N3O2',
    'Hcip': 'C7H12N3O2',
    'His': 'C6H7N3O',
    'Hisp': 'C6H6N3O',
    'Hser': 'C4H7NO2',
    'Hserp': 'C4H6NO2',
    'Hx': 'C6H11',
    'Hyp': 'C5H7NO2',
    'Hypp': 'C5H6NO2',
    'Ile': 'C6H11NO',
    'Ivdde': 'C14H21O2',
    'Leu': 'C6H11NO',
    'Lys': 'C6H12N2O',
    'Lysp': 'C6H11N2O',
    'Mbh': 'C15H15O2',
    'Me': 'CH3',
    'Mebzl': 'C8H9',
    'Meobzl': 'C8H9O',
    'Met': 'C5H9NOS',
    'Mmt': 'C20H17O',
    'Mtc': 'C14H19O3S',
    'Mtr': 'C10H13O3S',
    'Mts': 'C9H11O2S',
    'Mtt': 'C20H17',
    'Nle': 'C6H11NO',
    'Npys': 'C5H3N2O2S',
    'Nva': 'C5H9NO',
    'Odmab': 'C20H26NO3',
    'Orn': 'C5H10N2O',
    'Ornp': 'C5H9N2O',
    'Pbf': 'C13H17O3S',
    'Pen': 'C5H9NOS',
    'Penp': 'C5H8NOS',
    'Ph': 'C6H5',
    'Phe': 'C9H9NO',
    'Phepcl': 'C9H8ClNO',
    'Phg': 'C8H7NO',
    'Pmc': 'C14H19O3S',
    'Ppa': 'C8H7O2',
    'Pro': 'C5H7NO',
    'Prop': 'C3H7',
    'Py': 'C5H5N',
    'Pyr': 'C5H5NO2',
    'Sar': 'C3H5NO',
    'Ser': 'C3H5NO2',
    'Serp': 'C3H4NO2',
    'Sta': 'C8H15NO2',
    'Stap': 'C8H14NO2',
    'Tacm': 'C6H12NO',
    'Tbdms': 'C6H15Si',
    'Tbu': 'C4H9',
    'Tbuo': 'C4H9O',
    'Tbuthio': 'C4H9S',
    'Tfa': 'C2F3O',
    'Thi': 'C7H7NOS',
    'Thr': 'C4H7NO2',
    'Thrp': 'C4H6NO2',
    'Thy': 'C5H6N2O2',
    'Tips': 'C9H21Si',
    'Tms': 'C3H9Si',
    'Tos': 'C7H7O2S',
    'Trp': 'C11H10N2O',
    'Trpp': 'C11H9N2O',
    'Trt': 'C19H15',
    'Tyr': 'C9H9NO2',
    'Tyrp': 'C9H8NO2',
    'Ura': 'C4H4N2O2',
    'Val': 'C5H9NO',
    'Valoh': 'C5H9NO2',
    'Valohp': 'C5H8NO2',
    'Xan': 'C13H9O',
}

# Amino acids - H2O
AMINOACIDS: dict[str, str] = {
    'G': 'C2H3NO',  # Glycine, Gly
    'P': 'C5H7NO',  # Proline, Pro
    'A': 'C3H5NO',  # Alanine, Ala
    'V': 'C5H9NO',  # Valine, Val
    'L': 'C6H11NO',  # Leucine, Leu
    'I': 'C6H11NO',  # Isoleucine, Ile
    'M': 'C5H9NOS',  # Methionine, Met
    'C': 'C3H5NOS',  # Cysteine, Cys
    'F': 'C9H9NO',  # Phenylalanine, Phe
    'Y': 'C9H9NO2',  # Tyrosine, Tyr
    'W': 'C11H10N2O',  # Tryptophan, Trp
    'H': 'C6H7N3O',  # Histidine, His
    'K': 'C6H12N2O',  # Lysine, Lys
    'R': 'C6H12N4O',  # Arginine, Arg
    'Q': 'C5H8N2O2',  # Glutamine, Gln
    'N': 'C4H6N2O2',  # Asparagine, Asn
    'E': 'C5H7NO3',  # Glutamic Acid, Glu
    'D': 'C4H5NO3',  # Aspartic Acid, Asp
    'S': 'C3H5NO2',  # Serine, Ser
    'T': 'C4H7NO2',  # Threonine, Thr
}

# Deoxynucleoside monophosphates - H2O
DEOXYNUCLEOTIDES: dict[str, str] = {
    'A': 'C10H12N5O5P',
    'T': 'C10H13N2O7P',
    'C': 'C9H12N3O6P',
    'G': 'C10H12N5O6P',
}

# Nucleoside monophosphates - H2O
NUCLEOTIDES: dict[str, str] = {
    'A': 'C10H12N5O6P',
    'U': 'C9H11N2O8P',
    'C': 'C9H12N3O7P',
    'G': 'C10H12N5O7P',
}

DEOXYNUCLEOTIDE_COMPLEMENTS: dict[str, str] = {
    'A': 'T',
    'T': 'A',
    'C': 'G',
    'G': 'C',
}

NUCLEOTIDE_COMPLEMENTS: dict[str, str] = {
    'A': 'U',
    'U': 'A',
    'C': 'G',
    'G': 'C',
}

# Formula preprocessors
PREPROCESSORS: dict[str, Callable[[str], str]] = {
    'peptide': from_peptide,
    'ssdna': lambda x: from_oligo(x, 'ssdna'),
    'dsdna': lambda x: from_oligo(x, 'dsdna'),
    'ssrna': lambda x: from_oligo(x, 'ssrna'),
    'dsrna': lambda x: from_oligo(x, 'dsrna'),
}


def test(verbose: bool = False) -> None:
    """Test module.

    Parameters:
        verbose: Print status of testing.

    """
    formula: str
    empirical: str
    mass: float

    f = Formula('EtOH')
    assert f.formula == 'C2H6O'
    assert f.empirical == 'C2H6O'
    assert f.mass == 46.068531
    assert f.atoms == 9
    assert f.gcd == 1
    assert f.isotope.mass == 46.041864812949996
    assert f.isotope.massnumber == 46
    assert f.isotope.abundance == 0.9756627354527866
    assert f.composition().astuple() == (
        ('C', 2, 24.02148, 0.5214292593788155),
        ('H', 6, 6.047646, 0.13127499116479316),
        ('O', 1, 15.999405, 0.34729574945639136),
    )
    spectrum = f.spectrum(min_fraction=1e-9, min_intensity=1e-9)
    assert spectrum.asdict() == {
        46: (46.04186481295, 0.9756627354527866, 100.0, 46.04186481295),
        47: (
            47.04532293299781,
            0.022149945780234485,
            2.2702461593918635,
            47.04532293299781,
        ),
        48: (
            48.04629172951169,
            0.002142167346900965,
            0.21956023009393977,
            48.04629172951169,
        ),
        49: (
            49.04956894435918,
            4.4886295288353736e-05,
            0.004600595437061851,
            49.04956894435918,
        ),
        50: (
            50.05315154462678,
            2.6452603656753454e-07,
            2.711244643824313e-05,
            50.05315154462678,
        ),
        51: (
            51.05909655531064,
            1.6186163397759485e-10,
            1.6589916586542372e-08,
            51.05909655531064,
        ),
    }, spectrum.asdict()

    assert abs(spectrum.mean - 46.06852122027406) < 1e-9, spectrum.mean
    assert spectrum.peak.astuple() == (
        46,
        46.04186481295,
        0.9756627354527866,
        100.0,
        46.04186481295,
    ), spectrum.peak.astuple()

    # these formulas should pass
    for formula, empirical, mass in [
        (''.join(e.symbol for e in ELEMENTS), '', 14693.181589000998),
        ('12C', '[12C]', 12.0),
        ('12CC', 'C[12C]', 24.0107),
        ('SO4_2-', '[O4S]2-', 96.06351715981813),
        ('[CHNOP[13C]]2-', '[C[13C]HNOP]2-', 87.003),
        ('[CHNOP[13C]]_2-', '[C[13C]HNOP]2-', 87.003),
        ('CHNOP[13C]-2', '[C[13C]HNOP]2-', 87.003),
        ('Co(Bpy)(CO)4', '', 327.16),
        ('CH3CH2Cl', 'C2H5Cl', 64.5147),
        ('C1000H1000', 'CH', 13018.68),
        ('Ru2(CO)8', 'C4O4Ru', 426.2232),
        ('RuClH(CO)(PPh3)3', 'C55H46ClOP3Ru', 952.41392),
        ('PhSiMe3', 'C9H14Si', 150.29566),
        ('Ph(CO)C(CH3)3', 'C11H14O', 162.23156),
        ('HGlyGluTyrOH', 'C16H21N3O7', 367.35864),
        ('HCysTyrIleGlnAsnCysProLeuNH2', 'C41H65N11O11S2', 952.1519),
        ('CGCGAATTCGCG', 'C116H148N46O73P12', 3726.4),
        ('MDRGEQGLLK', 'C47H83N15O16S', 1146.3),
        ('CDCl3', 'C[2H]Cl3', 120.384),
        ('[13C]Cl4', '[13C]Cl4', 154.8153),
        ('C5(PhBu(EtCHBr)2)3', 'C53H78Br6', 1194.626),
        ('AgCuRu4(H)2[CO]12{PPh3}2', 'C48H32AgCuO12P2Ru4', 1438.4022),
        ('PhNH2.HCl', 'C6H8ClN', 129.5892),
        ('NH3.BF3', 'BF3H3N', 84.8357),
        ('CuSO4.5H2O', 'CuH10O9S', 249.68),
        ('5*H2O+CuSO4', 'CuH10O9S', 249.68),
        ('5*H2O', 'H2O', 90.076435),
        (
            'HCysp(Trt)Tyrp(Tbu)IleGlnp(Trt)Asnp(Trt)ProLeuGlyNH2',
            'C101H113N11O11S',
            1689.13532,
        ),
    ]:
        if verbose:
            print(f'Trying Formula({formula!r}) ...', end='')
        try:
            f = Formula(formula)
            f.empirical
            f.mass
            f.spectrum()
        except FormulaError as exc:
            print('Error:', exc)
            continue
        if empirical and f.empirical != empirical:
            print(
                f'Failure for {formula!r}:\n    Expected {empirical!r}, '
                f'got {f.empirical!r}:'
            )
            continue
        if mass and abs(f.mass - mass) > 0.1:
            print(
                f'Failure for {formula!r}:\n    Expected {mass}, got {f.mass}'
            )
            continue
        if verbose:
            print('ok')

    # these formulas are expected to fail
    for formula in [
        '',
        '()',
        '2',
        'a',
        '(a)',
        'C:H',
        'H:',
        'C[H',
        'H)2',
        'A',
        'Aa',
        '2lC',
        '1C',
        '[11C]',
        'H0',
        '()0',
        '(H)0C',
        'Ox: 0.26, 30Si: 0.74',
        'H^++',
        '[CHNOP[13C]]__2-',
        'O2_-2',
        'C+a',
    ]:
        if verbose:
            print(f'Trying Formula({formula!r}) ...', end='')
        try:
            empirical = Formula(formula).empirical
        except FormulaError as exc:
            if verbose:
                print('ok\nExpected error:', exc)
        else:
            print(
                f'Failure expected for {formula!r}, got '
                f'{Formula(formula).formula!r}'
            )

    from molmass import PROTON

    assert abs(Formula('[1H]+').mass - PROTON.mass) < 1e-7


def main(argv: list[str] | None = None, /) -> int:
    """Command line usage main function.

    Parameters:
        argv: Command line arguments.

    """
    if argv is None:
        argv = sys.argv

    import optparse

    def search_doc(r: str, d: str) -> str:
        if not __doc__:
            return d
        match = re.search(r, __doc__)
        if match is None:
            return d
        return match.group(1)

    parser = optparse.OptionParser(
        usage='usage: %prog [options] formula',
        description=search_doc('\n\n([^|]*?)\n\n', ''),
        version='%prog ' + search_doc(':Version: (.*)', 'Unknown'),
        prog='molmass',
    )
    opt = parser.add_option
    opt(
        '--web',
        dest='web',
        action='store_true',
        default=False,
        help='start web application and open it in a web browser',
    )
    opt(
        '--url',
        dest='url',
        default=None,
        help='URL to run web application',
    )
    opt(
        '--nobrowser',
        dest='nobrowser',
        action='store_true',
        default=False,
        help='do not open web browser',
    )
    opt(
        '--test',
        dest='test',
        action='store_true',
        default=False,
        help='test the module',
    )
    opt(
        '--doctest',
        dest='doctest',
        action='store_true',
        default=False,
        help='run the internal tests',
    )
    opt('-v', '--verbose', dest='verbose', action='store_true', default=False)

    settings, formula_list = parser.parse_args()

    if settings.web:
        try:
            from .web import main as web_main
        except ImportError:
            from web import main as web_main  # type: ignore

        if formula_list:
            form = {'q': ''.join(formula_list)}
        else:
            form = None

        return web_main(
            url=settings.url, open_browser=not settings.nobrowser, form=form
        )
    if settings.doctest:
        import doctest

        try:
            import molmass.molmass as m
        except ImportError:
            m = None  # type: ignore
        doctest.testmod(m, optionflags=doctest.ELLIPSIS)
        return 0
    if settings.test:
        test(settings.verbose)
        return 0
    if formula_list:
        formula = ''.join(formula_list)
    else:
        parser.error('no formula specified')

    try:
        results = analyze(formula)
    except Exception as exc:
        print('\nError: \n  ', exc, sep='')
        raise exc
    print('\n', results, sep='')

    return 0


if '--doctest' in sys.argv:
    # enable doctest for @cached_property
    __test__ = {  # noqa
        'Formula._elements': Formula._elements,
        'Formula.formula': Formula.formula,
        'Formula.empirical': Formula.empirical,
        'Formula.atoms': Formula.atoms,
        'Formula.gcd': Formula.gcd,
        'Formula.mass': Formula.mass,
        'Formula.isotope': Formula.isotope,
        'Spectrum.range': Spectrum.range,
        'Spectrum.peak': Spectrum.peak,
        'Spectrum.mean': Spectrum.mean,
        'Composition.total': Composition.total,
    }

if __name__ == '__main__':
    sys.exit(main())



# molmass/elements_descriptions.py

"""Set the description attribute on all elements."""

try:
    from .elements import ELEMENTS
except ImportError:
    from elements import ELEMENTS  # type: ignore

ELEMENTS['H'].description = (
    "Colourless, odourless gaseous chemical element. Lightest and most "
    "abundant element in the universe. Present in water and in all organic "
    "compounds. Chemically reacts with most elements. Discovered by Henry "
    "Cavendish in 1776."
)
ELEMENTS['He'].description = (
    "Colourless, odourless gaseous nonmetallic element. Belongs to group "
    "18 of the periodic table. Lowest boiling point of all elements and "
    "can only be solidified under pressure. Chemically inert, no known "
    "compounds. Discovered in the solar spectrum in 1868 by Lockyer."
)
ELEMENTS['Li'].description = (
    "Socket silvery metal. First member of group 1 of the periodic table. "
    "Lithium salts are used in psychomedicine."
)
ELEMENTS['Be'].description = (
    "Grey metallic element of group 2 of the periodic table. Is toxic and "
    "can cause severe lung diseases and dermatitis. Shows high covalent "
    "character. It was isolated independently by F. Wohler and A.A. Bussy "
    "in 1828."
)
ELEMENTS['B'].description = (
    "An element of group 13 of the periodic table. There are two "
    "allotropes, amorphous boron is a brown power, but metallic boron is "
    "black. The metallic form is hard (9.3 on Mohs' scale) and a bad "
    "conductor in room temperatures. It is never found free in nature. "
    "Boron-10 is used in nuclear reactor control rods and shields. It was "
    "discovered in 1808 by Sir Humphry Davy and by J.L. Gay-Lussac and "
    "L.J. Thenard."
)
ELEMENTS['C'].description = (
    "Carbon is a member of group 14 of the periodic table. It has three "
    "allotropic forms of it, diamonds, graphite and fullerite. Carbon-14 "
    "is commonly used in radioactive dating. Carbon occurs in all organic "
    "life and is the basis of organic chemistry. Carbon has the "
    "interesting chemical property of being able to bond with itself, and "
    "a wide variety of other elements."
)
ELEMENTS['N'].description = (
    "Colourless, gaseous element which belongs to group 15 of the "
    "periodic table. Constitutes ~78% of the atmosphere and is an "
    "essential part of the ecosystem. Nitrogen for industrial purposes is "
    "acquired by the fractional distillation of liquid air. Chemically "
    "inactive, reactive generally only at high temperatures or in "
    "electrical discharges. It was discovered in 1772 by D. Rutherford."
)
ELEMENTS['O'].description = (
    "A colourless, odourless gaseous element belonging to group 16 of the "
    "periodic table. It is the most abundant element present in the "
    "earth's crust. It also makes up 20.8% of the Earth's atmosphere. For "
    "industrial purposes, it is separated from liquid air by fractional "
    "distillation. It is used in high temperature welding, and in "
    "breathing. It commonly comes in the form of Oxygen, but is found as "
    "Ozone in the upper atmosphere. It was discovered by Priestley in "
    "1774."
)
ELEMENTS['F'].description = (
    "A poisonous pale yellow gaseous element belonging to group 17 of the "
    "periodic table (The halogens). It is the most chemically reactive and "
    "electronegative element. It is highly dangerous, causing severe "
    "chemical burns on contact with flesh. Fluorine was identified by "
    "Scheele in 1771 and first isolated by Moissan in 1886."
)
ELEMENTS['Ne'].description = (
    "Colourless gaseous element of group 18 on the periodic table (noble "
    "gases). Neon occurs in the atmosphere, and comprises 0.0018% of the "
    "volume of the atmosphere. It has a distinct reddish glow when used in "
    "discharge tubes and neon based lamps. It forms almost no chemical "
    "compounds. Neon was discovered in 1898 by Sir William Ramsey and M.W. "
    "Travers."
)
ELEMENTS['Na'].description = (
    "Soft silvery reactive element belonging to group 1 of the periodic "
    "table (alkali metals). It is highly reactive, oxidizing in air and "
    "reacting violently with water, forcing it to be kept under oil. It "
    "was first isolated by Humphrey Davy in 1807."
)
ELEMENTS['Mg'].description = (
    "Silvery metallic element belonging to group 2 of the periodic table "
    "(alkaline-earth metals). It is essential for living organisms, and is "
    "used in a number of light alloys. Chemically very reactive, it forms "
    "a protective oxide coating when exposed to air and burns with an "
    "intense white flame. It also reacts with sulphur, nitrogen and the "
    "halogens. First isolated by Bussy in 1828."
)
ELEMENTS['Al'].description = (
    "Silvery-white lustrous metallic element of group 3 of the periodic "
    "table. Highly reactive but protected by a thin transparent layer of "
    "the oxide which quickly forms in air. There are many alloys of "
    "aluminum, as well as a good number of industrial uses. Makes up 8.1% "
    "of the Earth's crust, by weight. Isolated in 1825 by H.C. Oersted."
)
ELEMENTS['Si'].description = (
    "Metalloid element belonging to group 14 of the periodic table. It is "
    "the second most abundant element in the Earth's crust, making up "
    "25.7% of it by weight. Chemically less reactive than carbon. First "
    "identified by Lavoisier in 1787 and first isolated in 1823 by "
    "Berzelius."
)
ELEMENTS['P'].description = (
    "Non-metallic element belonging to group 15 of the periodic table. "
    "Has a multiple allotropic forms. Essential element for living "
    "organisms. It was discovered by Brandt in 1669."
)
ELEMENTS['S'].description = (
    "Yellow, nonmetallic element belonging to group 16 of the periodic "
    "table. It is an essential element in living organisms, needed in the "
    "amino acids cysteine and methionine, and hence in many proteins. "
    "Absorbed by plants from the soil as sulphate ion."
)
ELEMENTS['Cl'].description = (
    "Halogen element. Poisonous greenish-yellow gas. Occurs widely in "
    "nature as sodium chloride in seawater. Reacts directly with many "
    "elements and compounds, strong oxidizing agent. Discovered by Karl "
    "Scheele in 1774. Humphrey David confirmed it as an element in 1810."
)
ELEMENTS['Ar'].description = (
    "Monatomic noble gas. Makes up 0.93% of the air. Colourless, "
    "odorless. Is inert and has no true compounds. Lord Rayleigh and Sir "
    "william Ramsey identified argon in 1894."
)
ELEMENTS['K'].description = (
    "Soft silvery metallic element belonging to group 1 of the periodic "
    "table (alkali metals). Occurs naturally in seawater and a many "
    "minerals. Highly reactive, chemically, it resembles sodium in its "
    "behavior and compounds. Discovered by Sir Humphry Davy in 1807."
)
ELEMENTS['Ca'].description = (
    "Soft grey metallic element belonging to group 2 of the periodic "
    "table. Used a reducing agent in the extraction of thorium, zirconium "
    "and uranium. Essential element for living organisms."
)
ELEMENTS['Sc'].description = (
    "Rare soft silvery metallic element belonging to group 3 of the "
    "periodic table. There are ten isotopes, nine of which are radioactive "
    "and have short half-lives. Predicted in 1869 by Mendeleev, isolated "
    "by Nilson in 1879."
)
ELEMENTS['Ti'].description = (
    "White metallic transition element. Occurs in numerous minerals. Used "
    "in strong, light corrosion-resistant alloys. Forms a passive oxide "
    "coating when exposed to air. First discovered by Gregor in 1789."
)
ELEMENTS['V'].description = (
    "Soft and ductile, bright white metal. Good resistance to corrosion "
    "by alkalis, sulphuric and hydrochloric acid. It oxidizes readily "
    "about 933K. There are two naturally occurring isotopes of vanadium, "
    "and 5 radioisotopes, V-49 having the longest half-life at 337 days. "
    "Vanadium has nuclear applications, the foil is used in cladding "
    "titanium to steel, and vanadium-gallium tape is used to produce a "
    "superconductive magnet. Originally discovered by Andres Manuel del "
    "Rio of Mexico City in 1801. His discovery went unheeded, however, and "
    "in 1820, Nils Gabriel Sefstron of Sweden rediscovered it. Metallic "
    "vanadium was isolated by Henry Enfield Roscoe in 1867. The name "
    "vanadium comes from Vanadis, a goddess of Scandinavian mythology. "
    "Silvery-white metallic transition element. Vanadium is essential to "
    "Ascidians. Rats and chickens are also known to require it. Metal "
    "powder is a fire hazard, and vanadium compounds should be considered "
    "highly toxic. May cause lung cancer if inhaled."
)
ELEMENTS['Cr'].description = (
    "Hard silvery transition element. Used in decorative electroplating. "
    "Discovered in 1797 by Vauquelin."
)
ELEMENTS['Mn'].description = (
    "Grey brittle metallic transition element. Rather electropositive, "
    "combines with some non-metals when heated. Discovered in 1774 by "
    "Scheele."
)
ELEMENTS['Fe'].description = (
    "Silvery malleable and ductile metallic transition element. Has nine "
    "isotopes and is the fourth most abundant element in the earth's "
    "crust. Required by living organisms as a trace element (used in "
    "hemoglobin in humans.) Quite reactive, oxidizes in moist air, "
    "displaces hydrogen from dilute acids and combines with nonmetallic "
    "elements."
)
ELEMENTS['Co'].description = (
    "Light grey transition element. Some meteorites contain small amounts "
    "of metallic cobalt. Generally alloyed for use. Mammals require small "
    "amounts of cobalt salts. Cobalt-60, an artificially produced "
    "radioactive isotope of Cobalt is an important radioactive tracer and "
    "cancer-treatment agent. Discovered by G. Brandt in 1737."
)
ELEMENTS['Ni'].description = (
    "Malleable ductile silvery metallic transition element. Discovered by "
    "A.F. Cronstedt in 1751."
)
ELEMENTS['Cu'].description = (
    "Red-brown transition element. Known by the Romans as 'cuprum.' "
    "Extracted and used for thousands of years. Malleable, ductile and an "
    "excellent conductor of heat and electricity. When in moist "
    "conditions, a greenish layer forms on the outside."
)
ELEMENTS['Zn'].description = (
    "Blue-white metallic element. Occurs in multiple compounds naturally. "
    "Five stable isotopes are six radioactive isotopes have been found. "
    "Chemically a reactive metal, combines with oxygen and other "
    "non-metals, reacts with dilute acids to release hydrogen."
)
ELEMENTS['Ga'].description = (
    "Soft silvery metallic element, belongs to group 13 of the periodic "
    "table. The two stable isotopes are Ga-69 and Ga-71. Eight radioactive "
    "isotopes are known, all having short half-lives. Gallium Arsenide is "
    "used as a semiconductor. Corrodes most other metals by diffusing into "
    "their lattice. First identified by Francois Lecoq de Boisbaudran in "
    "1875."
)
ELEMENTS['Ge'].description = (
    "Lustrous hard metalloid element, belongs to group 14 of the periodic "
    "table. Forms a large number of organometallic compounds. Predicted by "
    "Mendeleev in 1871, it was actually found in 1886 by Winkler."
)
ELEMENTS['As'].description = (
    "Metalloid element of group 15. There are three allotropes, yellow, "
    "black, and grey. Reacts with halogens, concentrated oxidizing acids "
    "and hot alkalis. Albertus Magnus is believed to have been the first "
    "to isolate the element in 1250."
)
ELEMENTS['Se'].description = (
    "Metalloid element, belongs to group 16 of the periodic table. "
    "Multiple allotropic forms exist. Chemically resembles sulphur. "
    "Discovered in 1817 by Jons J. Berzelius."
)
ELEMENTS['Br'].description = (
    "Halogen element. Red volatile liquid at room temperature. Its "
    "reactivity is somewhere between chlorine and iodine. Harmful to human "
    "tissue in a liquid state, the vapour irritates eyes and throat. "
    "Discovered in 1826 by Antoine Balard."
)
ELEMENTS['Kr'].description = (
    "Colorless gaseous element, belongs to the noble gases. Occurs in the "
    "air, 0.0001% by volume. It can be extracted from liquid air by "
    "fractional distillation. Generally not isolated, but used with other "
    "inert gases in fluorescent lamps. Five natural isotopes, and five "
    "radioactive isotopes. Kr-85, the most stable radioactive isotope, has "
    "a half-life of 10.76 years and is produced in fission reactors. "
    "Practically inert, though known to form compounds with Fluorine."
)
ELEMENTS['Rb'].description = (
    "Soft silvery metallic element, belongs to group 1 of the periodic "
    "table. Rb-97, the naturally occurring isotope, is radioactive. It is "
    "highly reactive, with properties similar to other elements in group "
    "1, like igniting spontaneously in air. Discovered spectroscopically "
    "in 1861 by W. Bunsen and G.R. Kirchoff."
)
ELEMENTS['Sr'].description = (
    "Soft yellowish metallic element, belongs to group 2 of the periodic "
    "table. Highly reactive chemically. Sr-90 is present in radioactive "
    "fallout and has a half-life of 28 years. Discovered in 1798 by "
    "Klaproth and Hope, isolated in 1808 by Humphry Davy."
)
ELEMENTS['Y'].description = (
    "Silvery-grey metallic element of group 3 on the periodic table. "
    "Found in uranium ores. The only natural isotope is Y-89, there are 14 "
    "other artificial isotopes. Chemically resembles the lanthanoids. "
    "Stable in the air below 400 degrees, celsius. Discovered in 1828 by "
    "Friedrich Wohler."
)
ELEMENTS['Zr'].description = (
    "Grey-white metallic transition element. Five natural isotopes and "
    "six radioactive isotopes are known. Used in nuclear reactors for a "
    "Neutron absorber. Discovered in 1789 by Martin Klaproth, isolated in "
    "1824 by Berzelius."
)
ELEMENTS['Nb'].description = (
    "Soft, ductile grey-blue metallic transition element. Used in special "
    "steels and in welded joints to increase strength. Combines with "
    "halogens and oxidizes in air at 200 degrees celsius. Discovered by "
    "Charles Hatchett in 1801 and isolated by Blomstrand in 1864. Called "
    "Columbium originally."
)
ELEMENTS['Mo'].description = (
    "Silvery-white, hard metallic transition element. It is chemically "
    "unreactive and is not affected by most acids. It oxidizes at high "
    "temperatures. There are seven natural isotopes, and four "
    "radioisotopes, Mo-93 being the most stable with a half-life of 3500 "
    "years. Molybdenum is used in almost all high-strength steels, it has "
    "nuclear applications, and is a catalyst in petroleum refining. "
    "Discovered in 1778 by Carl Welhelm Scheele of Sweden. Impure metal "
    "was prepared in 1782 by Peter Jacob Hjelm. The name comes from the "
    "Greek word molybdos which means lead. Trace amounts of molybdenum are "
    "required for all known forms of life. All molybdenum compounds should "
    "be considered highly toxic, and will also cause severe birth defects."
)
ELEMENTS['Tc'].description = (
    "Radioactive metallic transition element. Can be detected in some "
    "stars and the fission products of uranium. First made by Perrier and "
    "Segre by bombarding molybdenum with deutrons, giving them Tc-97. "
    "Tc-99 is the most stable isotope with a half-life of 2.6*10^6 years. "
    "Sixteen isotopes are known. Organic technetium compounds are used in "
    "bone imaging. Chemical properties are intermediate between rhenium "
    "and manganese."
)
ELEMENTS['Ru'].description = (
    "Hard white metallic transition element. Found with platinum, used as "
    "a catalyst in some platinum alloys. Dissolves in fused alkalis, and "
    "is not attacked by acids. Reacts with halogens and oxygen at high "
    "temperatures. Isolated in 1844 by K.K. Klaus."
)
ELEMENTS['Rh'].description = (
    "Silvery white metallic transition element. Found with platinum and "
    "used in some platinum alloys. Not attacked by acids, dissolves only "
    "in aqua regia. Discovered in 1803 by W.H. Wollaston."
)
ELEMENTS['Pd'].description = (
    "Soft white ductile transition element. Found with some copper and "
    "nickel ores. Does not react with oxygen at normal temperatures. "
    "Dissolves slowly in hydrochloric acid. Discovered in 1803 by W.H. "
    "Wollaston."
)
ELEMENTS['Ag'].description = (
    "White lustrous soft metallic transition element. Found in both its "
    "elemental form and in minerals. Used in jewellery, tableware and so "
    "on. Less reactive than silver, chemically."
)
ELEMENTS['Cd'].description = (
    "Soft bluish metal belonging to group 12 of the periodic table. "
    "Extremely toxic even in low concentrations. Chemically similar to "
    "zinc, but lends itself to more complex compounds. Discovered in 1817 "
    "by F. Stromeyer."
)
ELEMENTS['In'].description = (
    "Soft silvery element belonging to group 13 of the periodic table. "
    "The most common natural isotope is In-115, which has a half-life of "
    "6*10^4 years. Five other radioisotopes exist. Discovered in 1863 by "
    "Reich and Richter."
)
ELEMENTS['Sn'].description = (
    "Silvery malleable metallic element belonging to group 14 of the "
    "periodic table. Twenty-six isotopes are known, five of which are "
    "radioactive. Chemically reactive. Combines directly with chlorine and "
    "oxygen and displaces hydrogen from dilute acids."
)
ELEMENTS['Sb'].description = (
    "Element of group 15. Multiple allotropic forms. The stable form of "
    "antimony is a blue-white metal. Yellow and black antimony are "
    "unstable non-metals. Used in flame-proofing, paints, ceramics, "
    "enamels, and rubber. Attacked by oxidizing acids and halogens. First "
    "reported by Tholden in 1450."
)
ELEMENTS['Te'].description = (
    "Silvery metalloid element of group 16. Eight natural isotopes, nine "
    "radioactive isotopes. Used in semiconductors and to a degree in some "
    "steels. Chemistry is similar to Sulphur. Discovered in 1782 by Franz "
    "Miller."
)
ELEMENTS['I'].description = (
    "Dark violet nonmetallic element, belongs to group 17 of the periodic "
    "table. Insoluble in water. Required as a trace element for living "
    "organisms. One stable isotope, I-127 exists, in addition to fourteen "
    "radioactive isotopes. Chemically the least reactive of the halogens, "
    "and the most electropositive metallic halogen. Discovered in 1812 by "
    "Courtois."
)
ELEMENTS['Xe'].description = (
    "Colourless, odourless gas belonging to group 18 on the periodic "
    "table (the noble gases.) Nine natural isotopes and seven radioactive "
    "isotopes are known. Xenon was part of the first noble-gas compound "
    "synthesized. Several others involving Xenon have been found since "
    "then. Xenon was discovered by Ramsey and Travers in 1898."
)
ELEMENTS['Cs'].description = (
    "Soft silvery-white metallic element belonging to group 1 of the "
    "periodic table. One of the three metals which are liquid at room "
    "temperature. Cs-133 is the natural, and only stable, isotope. Fifteen "
    "other radioisotopes exist. Caesium reacts explosively with cold "
    "water, and ice at temperatures above 157K. Caesium hydroxide is the "
    "strongest base known. Caesium is the most electropositive, most "
    "alkaline and has the least ionization potential of all the elements. "
    "Known uses include the basis of atomic clocks, catalyst for the "
    "hydrogenation of some organic compounds, and in photoelectric cells. "
    "Caesium was discovered by Gustav Kirchoff and Robert Bunsen in "
    "Germany in 1860 spectroscopically. Its identification was based upon "
    "the bright blue lines in its spectrum. The name comes from the latin "
    "word caesius, which means sky blue. Caesium should be considered "
    "highly toxic. Some of the radioisotopes are even more toxic."
)
ELEMENTS['Ba'].description = (
    "Silvery-white reactive element, belonging to group 2 of the periodic "
    "table. Soluble barium compounds are extremely poisonous. Identified "
    "in 1774 by Karl Scheele and extracted in 1808 by Humphry Davy."
)
ELEMENTS['La'].description = (
    "(From the Greek word lanthanein, to line hidden) Silvery metallic "
    "element belonging to group 3 of the periodic table and oft considered "
    "to be one of the lanthanoids. Found in some rare-earth minerals. "
    "Twenty-five natural isotopes exist. La-139 which is stable, and "
    "La-138 which has a half-life of 10^10 to 10^15 years. The other "
    "twenty-three isotopes are radioactive. It resembles the lanthanoids "
    "chemically. Lanthanum has a low to moderate level of toxicity, and "
    "should be handled with care. Discovered in 1839 by C.G. Mosander."
)
ELEMENTS['Ce'].description = (
    "Silvery metallic element, belongs to the lanthanoids. Four natural "
    "isotopes exist, and fifteen radioactive isotopes have been "
    "identified. Used in some rare-earth alloys. The oxidized form is used "
    "in the glass industry. Discovered by Martin .H. Klaproth in 1803."
)
ELEMENTS['Pr'].description = (
    "Soft silvery metallic element, belongs to the lanthanoids. Only "
    "natural isotope is Pr-141 which is not radioactive. Fourteen "
    "radioactive isotopes have been artificially produced. Used in "
    "rare-earth alloys. Discovered in 1885 by C.A. von Welsbach."
)
ELEMENTS['Nd'].description = (
    "Soft bright silvery metallic element, belongs to the lanthanoids. "
    "Seven natural isotopes, Nd-144 being the only radioactive one with a "
    "half-life of 10^10 to 10^15 years. Six artificial radioisotopes have "
    "been produced. The metal is used in glass works to color class a "
    "shade of violet-purple and make it dichroic. One of the more reactive "
    "rare-earth metals, quickly reacts with air. Used in some rare-earth "
    "alloys. Neodymium is used to color the glass used in welder's "
    "glasses. Neodymium is also used in very powerful, permanent magnets "
    "(Nd2Fe14B). Discovered by Carl F. Auer von Welsbach in Austria in "
    "1885 by separating didymium into its elemental components "
    "Praseodymium and neodymium. The name comes from the Greek words 'neos "
    "didymos' which means 'new twin'. Neodymium should be considered "
    "highly toxic, however evidence would seem to show that it acts as "
    "little more than a skin and eye irritant. The dust however, presents "
    "a fire and explosion hazard."
)
ELEMENTS['Pm'].description = (
    "Soft silvery metallic element, belongs to the lanthanoids. Pm-147, "
    "the only natural isotope, is radioactive and has a half-life of 252 "
    "years. Eighteen radioisotopes have been produced, but all have very "
    "short half-lives. Found only in nuclear decay waste. Pm-147 is of "
    "interest as a beta-decay source, however Pm-146 and Pm-148 have to be "
    "removed from it first, as they generate gamma radiation. Discovered "
    "by J.A. Marinsky, L.E. Glendenin and C.D. Coryell in 1947."
)
ELEMENTS['Sm'].description = (
    "Soft silvery metallic element, belongs to the lanthanoids. Seven "
    "natural isotopes, Sm-147 is the only radioisotope, and has a "
    "half-life of 2.5*10^11 years. Used for making special alloys needed "
    "in the production of nuclear reactors. Also used as a neutron "
    "absorber. Small quantities of samarium oxide is used in special "
    "optical glasses. The largest use of the element is its ferromagnetic "
    "alloy which produces permanent magnets that are five times stronger "
    "than magnets produced by any other material. Discovered by Francois "
    "Lecoq de Boisbaudran in 1879."
)
ELEMENTS['Eu'].description = (
    "Soft silvery metallic element belonging to the lanthanoids. Eu-151 "
    "and Eu-153 are the only two stable isotopes, both of which are "
    "Neutron absorbers. Discovered in 1889 by Sir William Crookes."
)
ELEMENTS['Gd'].description = (
    "Soft silvery metallic element belonging to the lanthanoids. Seven "
    "natural, stable isotopes are known in addition to eleven artificial "
    "isotopes. Gd-155 and Gd-157 and the best neutron absorbers of all "
    "elements. Gadolinium compounds are used in electronics. Discovered by "
    "J.C.G Marignac in 1880."
)
ELEMENTS['Tb'].description = (
    "Silvery metallic element belonging to the lanthanoids. Tb-159 is the "
    "only stable isotope, there are seventeen artificial isotopes. "
    "Discovered by G.G. Mosander in 1843."
)
ELEMENTS['Dy'].description = (
    "Metallic with a bright silvery-white lustre. Dysprosium belongs to "
    "the lanthanoids. It is relatively stable in air at room temperatures, "
    "it will however dissolve in mineral acids, evolving hydrogen. It is "
    "found in from rare-earth minerals. There are seven natural isotopes "
    "of dysprosium, and eight radioisotopes, Dy-154 being the most stable "
    "with a half-life of 3*10^6 years. Dysprosium is used as a neutron "
    "absorber in nuclear fission reactions, and in compact disks. It was "
    "discovered by Paul Emile Lecoq de Boisbaudran in 1886 in France. Its "
    "name comes from the Greek word dysprositos, which means hard to "
    "obtain."
)
ELEMENTS['Ho'].description = (
    "Relatively soft and malleable silvery-white metallic element, which "
    "is stable in dry air at room temperature. It oxidizes in moist air "
    "and at high temperatures. It belongs to the lanthanoids. A rare-earth "
    "metal, it is found in the minerals monazite and gadolinite. It "
    "possesses unusual magnetic properties. One natural isotope, Ho-165 "
    "exists, six radioisotopes exist, the most stable being Ho-163 with a "
    "half-life of 4570 years. Holmium is used in some metal alloys, it is "
    "also said to stimulate the metabolism. Discovered by Per Theodor "
    "Cleve and J.L. Soret in Switzerland in 1879. The name homium comes "
    "from the Greek word Holmia which means Sweden. While all holmium "
    "compounds should be considered highly toxic, initial evidence seems "
    "to indicate that they do not pose much danger. The metal's dust "
    "however, is a fire hazard."
)
ELEMENTS['Er'].description = (
    "Soft silvery metallic element which belongs to the lanthanoids. Six "
    "natural isotopes that are stable. Twelve artificial isotopes are "
    "known. Used in nuclear technology as a neutron absorber. It is being "
    "investigated for other possible uses. Discovered by Carl G. Mosander "
    "in 1843."
)
ELEMENTS['Tm'].description = (
    "Soft grey metallic element that belongs to the lanthanoids. One "
    "natural isotope exists, Tm-169, and seventeen artificial isotopes "
    "have been produced. No known uses for the element. Discovered in 1879 "
    "by Per Theodor Cleve."
)
ELEMENTS['Yb'].description = (
    "Silvery metallic element of the lanthanoids. Seven natural isotopes "
    "and ten artificial isotopes are known. Used in certain steels. "
    "Discovered by J.D.G. Marignac in 1878."
)
ELEMENTS['Lu'].description = (
    "Silvery-white rare-earth metal which is relatively stable in air. It "
    "happens to be the most expensive rare-earth metal. Its found with "
    "almost all rare-earth metals, but is very difficult to separate from "
    "other elements. Least abundant of all natural elements. Used in metal "
    "alloys, and as a catalyst in various processes. There are two "
    "natural, stable isotopes, and seven radioisotopes, the most stable "
    "being Lu-174 with a half-life of 3.3 years. The separation of "
    "lutetium from Ytterbium was described by Georges Urbain in 1907. It "
    "was discovered at approximately the same time by Carl Auer von "
    "Welsbach. The name comes from the Greek word lutetia which means "
    "Paris."
)
ELEMENTS['Hf'].description = (
    "Silvery lustrous metallic transition element. Used in tungsten "
    "alloys in filaments and electrodes, also acts as a neutron absorber. "
    "First reported by Urbain in 1911, existence was finally established "
    "in 1923 by D. Coster, G.C. de Hevesy in 1923."
)
ELEMENTS['Ta'].description = (
    "Heavy blue-grey metallic transition element. Ta-181 is a stable "
    "isotope, and Ta-180 is a radioactive isotope, with a half-life in "
    "excess of 10^7 years. Used in surgery as it is unreactive. Forms a "
    "passive oxide layer in air. Identified in 1802 by Ekeberg and "
    "isolated in 1820 by Jons J. Berzelius."
)
ELEMENTS['W'].description = (
    "White or grey metallic transition element,formerly called Wolfram. "
    "Forms a protective oxide in air and can be oxidized at high "
    "temperature. First isolated by Jose and Fausto de Elhuyer in 1783."
)
ELEMENTS['Re'].description = (
    "Silvery-white metallic transition element. Obtained as a by-product "
    "of molybdenum refinement. Rhenium-molybdenum alloys are "
    "superconducting."
)
ELEMENTS['Os'].description = (
    "Hard blue-white metallic transition element. Found with platinum and "
    "used in some alloys with platinum and iridium."
)
ELEMENTS['Ir'].description = (
    "Very hard and brittle, silvery metallic transition element. It has a "
    "yellowish cast to it. Salts of iridium are highly colored. It is the "
    "most corrosion resistant metal known, not attacked by any acid, but "
    "is attacked by molten salts. There are two natural isotopes of "
    "iridium, and 4 radioisotopes, the most stable being Ir-192 with a "
    "half-life of 73.83 days. Ir-192 decays into Platinum, while the other "
    "radioisotopes decay into Osmium. Iridium is used in high temperature "
    "apparatus, electrical contacts, and as a hardening agent for "
    "platinumpy. Discovered in 1803 by Smithson Tennant in England. The "
    "name comes from the Greek word iris, which means rainbow. Iridium "
    "metal is generally non-toxic due to its relative unreactivity, but "
    "iridium compounds should be considered highly toxic."
)
ELEMENTS['Pt'].description = (
    "Attractive greyish-white metal. When pure, it is malleable and "
    "ductile. Does not oxidize in air, insoluble in hydrochloric and "
    "nitric acid. Corroded by halogens, cyandies, sulphur and alkalis. "
    "Hydrogen and Oxygen react explosively in the presence of platinumpy. "
    "There are six stable isotopes and three radioisotopes, the most "
    "stable being Pt-193 with a half-life of 60 years. Platinum is used in "
    "jewelry, laboratory equipment, electrical contacts, dentistry, and "
    "anti-pollution devices in cars. PtCl2(NH3)2 is used to treat some "
    "forms of cancer. Platinum-Cobalt alloys have magnetic properties. It "
    "is also used in the definition of the Standard Hydrogen Electrode. "
    "Discovered by Antonio de Ulloa in South America in 1735. The name "
    "comes from the Spanish word platina which means silver. Platinum "
    "metal is generally not a health concern due to its unreactivity, "
    "however platinum compounds should be considered highly toxic."
)
ELEMENTS['Au'].description = (
    "Gold is gold colored. It is the most malleable and ductile metal "
    "known. There is only one stable isotope of gold, and five "
    "radioisotopes of gold, Au-195 being the most stable with a half-life "
    "of 186 days. Gold is used as a monetary standard, in jewelry, "
    "dentistry, electronics. Au-198 is used in treating cancer and some "
    "other medical conditions. Gold has been known to exist as far back as "
    "2600 BC. Gold comes from the Anglo-Saxon word gold. Its symbol, Au, "
    "comes from the Latin word aurum, which means gold. Gold is not "
    "particularly toxic, however it is known to cause damage to the liver "
    "and kidneys in some."
)
ELEMENTS['Hg'].description = (
    "Heavy silvery liquid metallic element, belongs to the zinc group. "
    "Used in thermometers, barometers and other scientific apparatus. Less "
    "reactive than zinc and cadmium, does not displace hydrogen from "
    "acids. Forms a number of complexes and organomercury compounds."
)
ELEMENTS['Tl'].description = (
    "Pure, unreacted thallium appears silvery-white and exhibits a "
    "metallic lustre. Upon reacting with air, it begins to turn "
    "bluish-grey and looks like lead. It is very malleable, and can be cut "
    "with a knife. There are two stable isotopes, and four radioisotopes, "
    "Tl-204 being the most stable with a half-life of 3.78 years. Thallium "
    "sulphate was used as a rodenticide. Thallium sulphine's conductivity "
    "changes with exposure to infrared light, this gives it a use in "
    "infrared detectors. Discovered by Sir William Crookes via "
    "spectroscopy. Its name comes from the Greek word thallos, which means "
    "green twig. Thallium and its compounds are toxic and can cause "
    "cancer."
)
ELEMENTS['Pb'].description = (
    "Heavy dull grey ductile metallic element, belongs to group 14. Used "
    "in building construction, lead-place accumulators, bullets and shot, "
    "and is part of solder, pewter, bearing metals, type metals and "
    "fusible alloys."
)
ELEMENTS['Bi'].description = (
    "White crystalline metal with a pink tinge, belongs to group 15. Most "
    "diamagnetic of all metals and has the lowest thermal conductivity of "
    "all the elements except mercury. Lead-free bismuth compounds are used "
    "in cosmetics and medical procedures. Burns in the air and produces a "
    "blue flame. In 1753, C.G. Junine first demonstrated that it was "
    "different from lead."
)
ELEMENTS['Po'].description = (
    "Rare radioactive metallic element, belongs to group 16 of the "
    "periodic table. Over 30 known isotopes exist, the most of all "
    "elements. Po-209 has a half-life of 103 years. Possible uses in "
    "heating spacecraft. Discovered by Marie Curie in 1898 in a sample of "
    "pitchblende."
)
ELEMENTS['At'].description = (
    "Radioactive halogen element. Occurs naturally from uranium and "
    "thorium decay. At least 20 known isotopes. At-210, the most stable, "
    "has a half-life of 8.3 hours. Synthesized by nuclear bombardment in "
    "1940 by D.R. Corson, K.R. MacKenzie and E. Segre at the University of "
    "California."
)
ELEMENTS['Rn'].description = (
    "Colorless radioactive gaseous element, belongs to the noble gases. "
    "Of the twenty known isotopes, the most stable is Rn-222 with a "
    "half-life of 3.8 days. Formed by the radioactive decay of Radium-226. "
    "Radon itself decays into Polonium. Used in radiotherapy. As a noble "
    "gas, it is effectively inert, though radon fluoride has been "
    "synthesized. First isolated in 1908 by Ramsey and Gray."
)
ELEMENTS['Fr'].description = (
    "Radioactive element, belongs to group 1 of the periodic table. Found "
    "in uranium and thorium ores. The 22 known isotopes are all "
    "radioactive, with the most stable being Fr-223. Its existence was "
    "confirmed in 1939 by Marguerite Perey."
)
ELEMENTS['Ra'].description = (
    "Radioactive metallic transuranic element, belongs to group 2 of the "
    "periodic table. Most stable isotope, Ra-226 has a half-life of 1602 "
    "years, which decays into radon. Isolated from pitchblende in 1898 "
    "Marie and Pierre Curie."
)
ELEMENTS['Ac'].description = (
    "Silvery radioactive metallic element, belongs to group 3 of the "
    "periodic table. The most stable isotope, Ac-227, has a half-life of "
    "217 years. Ac-228 (half-life of 6.13 hours) also occurs in nature. "
    "There are 22 other artificial isotopes, all radioactive and having "
    "very short half-lives. Chemistry similar to lanthanumpy. Used as a "
    "source of alpha particles. Discovered by A. Debierne in 1899."
)
ELEMENTS['Th'].description = (
    "Grey radioactive metallic element. Belongs to actinoids. Found in "
    "monazite sand in Brazil, India and the US. Thorium-232 has a "
    "half-life of 1.39x10^10 years. Can be used as a nuclear fuel for "
    "breeder reactors. Thorium-232 captures slow Neutrons and breeds "
    "uranium-233. Discovered by Jons J. Berzelius in 1829."
)
ELEMENTS['Pa'].description = (
    "Radioactive metallic element, belongs to the actinoids. The most "
    "stable isotope, Pa-231 has a half-life of 2.43*10^4 years. At least "
    "10 other radioactive isotopes are known. No practical applications "
    "are known. Discovered in 1917 by Lise Meitner and Otto Hahn."
)
ELEMENTS['U'].description = (
    "White radioactive metallic element belonging to the actinoids. Three "
    "natural isotopes, U-238, U-235 and U-234. Uranium-235 is used as the "
    "fuel for nuclear reactors and weapons. Discovered by Martin H. "
    "Klaproth in 1789."
)
ELEMENTS['Np'].description = (
    "Radioactive metallic transuranic element, belongs to the actinoids. "
    "Np-237, the most stable isotope, has a half-life of 2.2*10^6 years "
    "and is a by product of nuclear reactors. The other known isotopes "
    "have mass numbers 229 through 236, and 238 through 241. Np-236 has a "
    "half-life of 5*10^3 years. First produced by Edwin M. McMillan and "
    "P.H. Abelson in 1940."
)
ELEMENTS['Pu'].description = (
    "Dense silvery radioactive metallic transuranic element, belongs to "
    "the actinoids. Pu-244 is the most stable isotope with a half-life of "
    "7.6*10^7 years. Thirteen isotopes are known. Pu-239 is the most "
    "important, it undergoes nuclear fission with slow neutrons and is "
    "hence important to nuclear weapons and reactors. Plutonium production "
    "is monitored down to the gram to prevent military misuse. First "
    "produced by Gleen T. Seaborg, Edwin M. McMillan, J.W. Kennedy and "
    "A.C. Wahl in 1940."
)
ELEMENTS['Am'].description = (
    "Radioactive metallic transuranic element, belongs to the actinoids. "
    "Ten known isotopes. Am-243 is the most stable isotope, with a "
    "half-life of 7.95*10^3 years. Discovered by Glenn T. Seaborg and "
    "associates in 1945, it was obtained by bombarding Uranium-238 with "
    "alpha particles."
)
ELEMENTS['Cm'].description = (
    "Radioactive metallic transuranic element. Belongs to actinoid "
    "series. Nine known isotopes, Cm-247 has a half-life of 1.64*10^7 "
    "years. First identified by Glenn T. Seaborg and associates in 1944, "
    "first produced by L.B. Werner and I. Perlman in 1947 by bombarding "
    "americium-241 with Neutrons. Named for Marie Curie."
)
ELEMENTS['Bk'].description = (
    "Radioactive metallic transuranic element. Belongs to actinoid "
    "series. Eight known isotopes, the most common Bk-247, has a half-life "
    "of 1.4*10^3 years. First produced by Glenn T. Seaborg and associates "
    "in 1949 by bombarding americium-241 with alpha particles."
)
ELEMENTS['Cf'].description = (
    "Radioactive metallic transuranic element. Belongs to actinoid "
    "series. Cf-251 has a half life of about 700 years. Nine isotopes are "
    "known. Cf-252 is an intense Neutron source, which makes it an intense "
    "Neutron source and gives it a use in Neutron activation analysis and "
    "a possible use as a radiation source in medicine. First produced by "
    "Glenn T. Seaborg and associates in 1950."
)
ELEMENTS['Es'].description = (
    "Appearance is unknown, however it is most probably metallic and "
    "silver or gray in color. Radioactive metallic transuranic element "
    "belonging to the actinoids. Es-254 has the longest half-life of the "
    "eleven known isotopes at 270 days. First identified by Albert Ghiorso "
    "and associates in the debris of the 1952 hydrogen bomb explosion. In "
    "1961 the first microgram quantities of Es-232 were separated. While "
    "einsteinium never exists naturally, if a sufficient amount was "
    "assembled, it would pose a radiation hazard."
)
ELEMENTS['Fm'].description = (
    "Radioactive metallic transuranic element, belongs to the actinoids. "
    "Ten known isotopes, most stable is Fm-257 with a half-life of 10 "
    "days. First identified by Albert Ghiorso and associates in the debris "
    "of the first hydrogen-bomb explosion in 1952."
)
ELEMENTS['Md'].description = (
    "Radioactive metallic transuranic element. Belongs to the actinoid "
    "series. Only known isotope, Md-256 has a half-life of 1.3 hours. "
    "First identified by Glenn T. Seaborg, Albert Ghiorso and associates "
    "in 1955. Alternative name Unnilunium has been proposed. Named after "
    "the 'inventor' of the periodic table, Dmitri Mendeleev."
)
ELEMENTS['No'].description = (
    "Radioactive metallic transuranic element, belongs to the actinoids. "
    "Seven known isotopes exist, the most stable being No-254 with a "
    "half-life of 255 seconds. First identified with certainty by Albert "
    "Ghiorso and Glenn T. Seaborg in 1966. Unnilbium has been proposed as "
    "an alternative name."
)
ELEMENTS['Lr'].description = (
    "Appearance unknown, however it is most likely silvery-white or grey "
    "and metallic. Lawrencium is a synthetic rare-earth metal. There are "
    "eight known radioisotopes, the most stable being Lr-262 with a "
    "half-life of 3.6 hours. Due to the short half-life of lawrencium, and "
    "its radioactivity, there are no known uses for it. Identified by "
    "Albert Ghiorso in 1961 at Berkeley. It was produced by bombarding "
    "californium with boron ions. The name is temporary IUPAC "
    "nomenclature, the origin of the name comes from Ernest O. Lawrence, "
    "the inventor of the cyclotron. If sufficient amounts of lawrencium "
    "were produced, it would pose a radiation hazard."
)
ELEMENTS['Rf'].description = (
    "Radioactive transactinide element. Expected to have similar chemical "
    "properties to those displayed by hafnium. Rf-260 was discovered by "
    "the Joint Nuclear Research Institute at Dubna (U.S.S.R.) in 1964. "
    "Researchers at Berkeley discovered Unq-257 and Unq-258 in 1964."
)
ELEMENTS['Db'].description = (
    "Also known as Hahnium, Ha. Radioactive transactinide element. "
    "Half-life of 1.6s. Discovered in 1970 by Berkeley researchers. So "
    "far, seven isotopes have been discovered."
)
ELEMENTS['Sg'].description = (
    "Half-life of 0.9 +/- 0.2 s. Discovered by the Joint Institute for "
    "Nuclear Research at Dubna (U.S.S.R.) in June of 1974. Its existence "
    "was confirmed by the Lawrence Berkeley Laboratory and Livermore "
    "National Laboratory in September of 1974."
)
ELEMENTS['Bh'].description = (
    "Radioactive transition metal. Half-life of approximately 1/500 s. "
    "Discovered by the Joint Institute for Nuclear Research at Dubna "
    "(U.S.S.R.) in 1976. Confirmed by West German physicists at the Heavy "
    "Ion Research Laboratory at Darmstadt."
)
ELEMENTS['Hs'].description = (
    "Radioactive transition metal first synthesized in 1984 by a German "
    "research team led by Peter Armbruster and Gottfried Muenzenberg at "
    "the Institute for Heavy Ion Research at Darmstadt."
)
ELEMENTS['Mt'].description = (
    "Half-life of approximately 5 ms. The creation of this element "
    "demonstrated that fusion techniques could indeed be used to make new, "
    "heavy nuclei. Made and identified by physicists of the Heavy Ion "
    "Research Laboratory, Darmstadt, West Germany in 1982. Named in honor "
    "of Lise Meitner, the Austrian physicist."
)

if __name__ == '__main__':
    try:
        from .elements import word_wrap
    except ImportError:
        from elements import word_wrap  # type: ignore

    for ele in ELEMENTS:
        descr = word_wrap(
            ele.description, linelen=70, indent=0, joinstr=' "\n    "'
        )
        print(
            f'ELEMENTS[{ele.symbol!r}].description = ' f'(\n    "{descr}"\n)'
        )
